#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jss.py

Classes representing a JSS, and its available API calls, represented
as JSSObjects.
"""
from __future__ import print_function

from __future__ import absolute_import
try:
    import cPickle  # Python 2.X
except ImportError:
    import _pickle as cPickle  # Python 3+

import sys
import gzip
import os
import platform
import re
import json
from xml.etree import ElementTree

sys.path.insert(0, '/Library/AutoPkg/JSSImporter')
import requests

try:
    from UserDict import UserDict  # Python 2.X
except ImportError:
    from collections import UserDict  # Python 3.3+

# from jss.nsurlsession_adapter import NSURLSessionAdapter
from .curl_adapter import CurlAdapter
from .auth import UAPIAuth
from .distribution_points import DistributionPoints
from .exceptions import GetError, PutError, PostError, DeleteError
from .jssobject import JSSObject
from . import jssobjects
from . import uapiobjects
from .queryset import QuerySet
from .tools import error_handler, quote_and_encode


# Pylint wants us to store our many attributes in a dictionary.
# However, to maintain backwards compatibility with the interface,
# we can't do that.
# pylint: disable=too-many-instance-attributes, too-many-public-methods
class JSS(object):
    """Represents a JAMF Software Server, with object search methods.

    Setup a JSS for making API requests.

    Provide either a JSSPrefs object OR specify url, user, and
    password to init. Other parameters are optional.

    Attributes:
        base_url (str): Full URL to the JSS, with port.
        user (str): API username.
        password (str): API password for user.
        token (str, optional): UAPI authorization token.
        repo_prefs (list): List of dicts of repository configuration data.
        verbose (bool): Whether to include extra output.
        session: "Session" used to make all HTTP requests through
            whichever network adapter is in use (default is the requests adapter).
        ssl_verify (bool): Whether to verify SSL traffic from the JSS
            is genuine.
        distribution_points (:obj:`DistributionPoints`): DistributionPoints
        max_age (int): Number of seconds cached object information
            should be kept before re-retrieving. Defaults to '-1'.

            Possible values:
                -1: Keep cached data forever, or until manually
                    retrieved with the `JSSObject.retrieve()` method.
                0: Retrieve data from data for every access.
                positive number: Number of seconds to keep.
        uapi (:obj:`UAPI`): Reference to the UAPI transport facade.
        api (:obj:`JSSAPI`): Reference to the classic API facade.

    Args:
        jss_prefs (:obj:`JSSPrefs`): A JSSPrefs object.
        url (str): String, full URL to a JSS, with port.
        user (str): API Username.
        password (str): API Password.

            repo_prefs: A list of dicts with repository names and
                passwords.
            repos: (Optional) List of file repositories dicts to
                    connect.
                repo dicts:
                    Each file-share distribution point requires:
                        name: String name of the distribution point.
                            Must match the value on the JSS.
                        password: String password for the read/write
                            user.
        repo_prefs (list): A list of dicts with repository names and passwords.

            repos: (Optional) List of file repositories dicts to
            connect.
                repo dicts:
                    Each file-share distribution point requires:
                        name: String name of the distribution point.
                            Must match the value on the JSS.
                        password: String password for the read/write
                            user.

                    This form uses the distributionpoints API call to
                    determine the remaining information. There is also
                    an explicit form; See distribution_points package
                    for more info

                    CDP and JDS types require one dict for the master,
                    with key:
                        type: String, either "CDP" or "JDS".

        ssl_verify (bool): Boolean whether to verify SSL traffic from the
            JSS is genuine.
        verbose (bool): Boolean whether to include extra output.
    """

    class UAPI(object):
        """This object represents the UAPI. All UAPI search methods will be attached here."""
        def __init__(self, jss, url=None):
            self.jss = jss
            self._base_url = url
            self.max_age = -1

        @property
        def base_url(self):
            """The URL to the Casper JSS, including port if needed."""
            return self._base_url

        @base_url.setter
        def base_url(self, url):
            """The URL to the Casper JSS, including port if needed."""
            # Remove the frequently included yet incorrect trailing slash.
            self._base_url = url.rstrip("/")

        @property
        def url(self):  # type: () -> str
            return "%s/%s" % (self.base_url, "uapi")

    class JSSAPI(object):
        """This object represents the XML API. All regular API search methods will be attached here."""
        def __init__(self, jss, url=None):
            self.jss = jss
            self._base_url = url

        @property
        def base_url(self):
            """The URL to the Casper JSS, including port if needed."""
            return self._base_url

        @base_url.setter
        def base_url(self, url):
            """The URL to the Casper JSS, including port if needed."""
            # Remove the frequently included yet incorrect trailing slash.
            self._base_url = url.rstrip("/")

        @property
        def url(self):  # type: () -> str
            return "%s/%s" % (self.base_url, "JSSResource")

    # pylint: disable=too-many-arguments
    def __init__(
        self, jss_prefs=None, url=None, user=None, password=None,
        repo_prefs=None, ssl_verify=True, verbose=False, **kwargs):

        if jss_prefs is not None:
            url = jss_prefs.url
            user = jss_prefs.user
            password = jss_prefs.password
            repo_prefs = jss_prefs.repos
            ssl_verify = jss_prefs.verify
            suppress_warnings = jss_prefs.suppress_warnings

        # TODO: This method currently accepts '**kwargs' to soften
        # the deprecation of the urllib warnings removal.

        self.base_url = url

        if 'adapter' in kwargs:
            self.session = kwargs['adapter']
        else:
            self.session = requests.session()

        self.user = user
        self.password = password
        self.token = None  # For uapi
        self.repo_prefs = repo_prefs if repo_prefs else []
        self.verbose = verbose
        self.ssl_verify = ssl_verify

        self.distribution_points = DistributionPoints(self)
        self.max_age = -1
        self.uapi = JSS.UAPI(self, url)
        self.api = JSS.JSSAPI(self, url)

    # pylint: disable=too-many-arguments

    @property
    def _url(self):
        """The URL to the Casper JSS API endpoints. Get only."""
        return self.api.url

    @property
    def base_url(self):
        """The URL to the Casper JSS, including port if needed."""
        return self._base_url

    @base_url.setter
    def base_url(self, url):
        """The URL to the Casper JSS, including port if needed."""
        # Remove the frequently included yet incorrect trailing slash.
        self._base_url = url.rstrip("/")

    @property
    def user(self):
        """Username used to connect to the Casper API"""
        return self.session.auth[0]

    @user.setter
    def user(self, value):
        """Username used to connect to the Casper API.

        Args:
            value (str): username.
        """
        auth = self.session.auth
        password = auth[1] if auth else ''
        self.session.auth = (value, password)

    @property
    def password(self):
        """Password used to connect to the Casper API"""
        return self.session.auth[1]

    @password.setter
    def password(self, value):
        """Password used to connect to the Casper API.

        Args:
            value (str): password.
        """
        auth = self.session.auth
        user = auth[0] if auth else ''
        self.session.auth = (user, value)

    @property
    def ssl_verify(self):
        """Boolean value for whether to verify SSL traffic is valid."""
        return self.session.verify

    @ssl_verify.setter
    def ssl_verify(self, value):
        """Boolean value for whether to verify SSL traffic is valid.

        Args:
            value: Boolean.
        """
        self.session.verify = value

    def mount_network_adapter(self, network_adapter):
        """Mount a network adapter that uses the Requests API.

        The existing user, password, and ssl_verify values are
        transferred to the new adapter.

        Args:
            network_adapter: A network adapter object that uses the
                basic Requests API. Included in python-jss are the
                CurlAdapter and RequestsAdapter.
        """
        auth = (self.user, self.password)
        ssl_verify = self.ssl_verify
        self.session = network_adapter
        self.user, self.password = auth
        self.ssl_verify = ssl_verify

    def get(self, url_path, headers=None, **kwargs):
        # type: (str) -> Union[ElementTree.Element, dict, bytes]
        """GET a url, handle errors, and return an etree.

        In general, it is better to use a higher level interface for
        API requests, like the search methods on this class, or the
        JSSObjects themselves.

        Args:
            url_path: String API endpoint path to GET (e.g. "packages")
            headers: [Optional] headers to add to the request

        Returns:
            ElementTree.Element for the XML returned from the JSS if the response was XML,
            dict if the response had Content-Type for json,
            bytes for anything else

        Raises:
            GetError if provided url_path has a >= 400 response, for
            example, if an object queried for does not exist (404). Will
            also raise GetError for bad XML.

            This behavior will change in the future for 404/Not Found
            to returning None.
        """
        request_url = os.path.join(self.base_url, quote_and_encode(url_path))
        if headers is None:  # Fall back to XML to support python-jss prior to addition of UAPI
            headers = {'Content-Type': 'text/xml', 'Accept': 'text/xml'}

        response = self.session.get(request_url, headers=headers, **kwargs)

        if response.status_code == 200 and self.verbose:
            print("GET %s: Success." % request_url)
        elif response.status_code >= 400:
            error_handler(GetError, response)

        if 'text/xml' in response.headers['content-type']:
            # ElementTree in python2 only accepts bytes.
            try:
                xmldata = ElementTree.fromstring(response.content)
                return xmldata
            except ElementTree.ParseError:
                raise GetError("Error Parsing XML:\n%s" % response.content)
        elif response.headers['content-type'].startswith('application/json'):
            return response.json()
        else:
            return response.content

    def post(self, url_path, data=None):
        # type: (str, Union[ElementTree.Element, dict]) -> str
        """POST an object to the JSS. For creating new objects only.

        The JSS responds with the new object's ID number if successful.

        Args:
            url_path: String API endpoint path to POST (e.g.
                "packages/id/0")
            data: xml.etree.ElementTree.Element with valid XML for the
                desired obj_class.

        Returns:
            str ID number of the newly created object.

        Raises:
            PostError if provided url_path has a >= 400 response.
        """
        # The JSS expects a post to ID 0 to create an object

        request_url = os.path.join(self.base_url, quote_and_encode(url_path))
        headers = {}

        if isinstance(data, ElementTree.Element):
            data = ElementTree.tostring(data, encoding='UTF-8')
            headers = {'Content-Type': 'text/xml', 'Accept': 'text/xml'}
        elif isinstance(data, dict):
            data = json.dumps(data)
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        elif isinstance(data, UserDict):
            data = json.dumps(data.data)
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        else:
            headers = {'Content-Type': 'application/octet-stream', 'Accept': '*/*'}

        response = self.session.post(request_url, data=data, headers=headers)

        if response.status_code == 201 and self.verbose:
            print("POST %s: Success" % request_url)
        elif response.status_code >= 400:
            error_handler(PostError, response)

        if 'text/xml' in response.headers['content-type']:
            id_ = re.search(r"<id>([0-9]+)</id>", response.content.decode("utf-8")).group(1)
        else:
            return response

        return id_

    def put(self, url_path, data):
        # type: (str, Union[ElementTree.Element, dict]) -> str
        """Update an existing object on the JSS.

        In general, it is better to use a higher level interface for
        updating objects, namely, making changes to a JSSObject subclass
        and then using its save method.

        Args:
            url_path: String API endpoint path to PUT, with ID (e.g.
                "packages/id/<object ID>")
            data: xml.etree.ElementTree.Element with valid XML for the
                desired obj_class.

        Raises:
            PutError if provided url_path has a >= 400 response.
        """
        request_url = os.path.join(self.base_url, quote_and_encode(url_path))
        headers = {}

        if isinstance(data, ElementTree.Element):
            data = ElementTree.tostring(data, encoding='UTF-8')
            headers = {'Content-Type': 'text/xml', 'Accept': 'text/xml'}
        elif isinstance(data, dict):
            data = json.dumps(data)
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        elif isinstance(data, UserDict):
            data = json.dumps(data.data)
            headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
        else:
            raise TypeError('Could not PUT unrecognised data type')

        response = self.session.put(request_url, data=data, headers=headers)

        if response.status_code == 201 and self.verbose:
            print("PUT %s: Success." % request_url)
        elif response.status_code >= 400:
            error_handler(PutError, response)

    def delete(self, url_path, data=None):
        # type: (str, Optional[Union[ElementTree.Element, dict]]) -> None
        """Delete an object from the JSS.

        In general, it is better to use a higher level interface for
        deleting objects, namely, using a JSSObject's delete method.

        Args:
            url_path: String API endpoint path to DEL, with ID (e.g.
                "packages/id/<object ID>")
            data: xml.etree.ElementTree.Element with valid XML for the
                desired obj_class. Most classes don't need this.

        Raises:
            DeleteError if provided url_path has a >= 400 response.
        """
        request_url = os.path.join(self.base_url, quote_and_encode(url_path))
        if data:
            data = ElementTree.tostring(data, encoding='UTF-8')
            response = self.session.delete(request_url, data=data,
                                           headers={'Content-Type': 'text/xml', 'Accept': 'text/xml'})
        else:
            response = self.session.delete(request_url)

        if response.status_code == 200 and self.verbose:
            print("DEL %s: Success." % request_url)
        elif response.status_code >= 400:
            error_handler(DeleteError, response)

    def retrieve_all(self):
        all_search_methods = [
            getattr(self, name) for name in jssobjects.__all__]

        all_objects = {}
        for method in all_search_methods:
            name = method.__name__
            try:
                result = method()
            except GetError as err:
                msg = "Unable to retrieve '{}'"
                if err.status_code == 401:
                    msg += "; permission error"
                print(msg.format(name))
                continue


            # Flat objects can go straight in.
            if isinstance(result, JSSObject):
                all_objects[name] = result.retrieve()
            # Container types need to be retrieved.
            else:
                try:
                    all_objects[name] = result.retrieve_all()
                except GetError:
                    # A failure to get means the object type has zero
                    # results.
                    print(name, " has no results! (GETERRROR)")
                    all_objects[name] = []

        return all_objects

    def pickle_all(self, path, compress=True):
        """Back up entire JSS to a Python Pickle.

        For each object type, retrieve all objects, and then pickle
        the entire smorgasbord. This will almost certainly take a long
        time!

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening, and `.gz`
                appended if compress=True and it's missing.
            compress (bool): Whether to gzip output. Default is True.
        """
        all_objects = self.retrieve_all()
        path = os.path.expanduser(path)
        gz_ext = ".gz"
        if compress and not path.endswith(gz_ext):
            path = path + gz_ext

        opener = gzip.open if compress else open
        with opener(path, 'wb') as file_handle:
            cPickle.Pickler(
                file_handle, cPickle.HIGHEST_PROTOCOL).dump(all_objects)

    @classmethod
    def from_pickle(cls, path):
        """Load all objects from pickle file and return as dict.

        The dict returned will have keys named the same as the
        JSSObject classes contained, and the values will be
        QuerySets of all full objects of that class (for example,
        the equivalent of my_jss.Computer().retrieve_all()).

        This method can potentially take a very long time!

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to load from.
                Path will have ~ expanded prior to opening.
        """
        gz_magic = "\x1f\x8b\x08"

        # Determine if file is gzipped.
        with open(os.path.expanduser(path), "rb") as pickle:
            pickle_magic = pickle.read(len(gz_magic))
            compressed = True if pickle_magic == gz_magic else False

        opener = gzip.open if compressed else open

        with opener(os.path.expanduser(path), "rb") as pickle:
            return cPickle.Unpickler(pickle).load()

    def write_all(self, path):
        """Back up entire JSS to XML file.

        For each object type, retrieve all objects, and then pickle
        the entire smorgasbord. This will almost certainly take a long
        time!

        Pickling is Python's method for serializing/deserializing
        Python objects. This allows you to save a fully functional
        JSSObject to disk, and then load it later, without having to
        retrieve it from the JSS.

        Args:
            path: String file path to the file you wish to (over)write.
                Path will have ~ expanded prior to opening.
        """
        all_objects = self.retrieve_all()

        with open(os.path.expanduser(path), "w") as ofile:
            root = ElementTree.Element("JSS")
            for obj_type, objects in all_objects.items():
                if objects:
                    sub_element = ElementTree.SubElement(root, obj_type)
                    sub_element.extend(objects)

            et = ElementTree.ElementTree(root)
            et.write(ofile, encoding="utf-8")

    def load_from_xml(self, path):
        """Load all objects from XML file and return as dict.

        The dict returned will have keys named the same as the
        JSSObject classes contained, and the values will be
        QuerySets of all full objects of that class (for example,
        the equivalent of my_jss.Computer().retrieve_all()).

        This method can potentially take a very long time!

        Args:
            path: String file path to the file you wish to load from.
                Path will have ~ expanded prior to opening.
        """
        with open(os.path.expanduser(path), "r") as ifile:
            et = ElementTree.parse(ifile)

        root = et.getroot()

        all_objects = {}
        for child in root:
            obj_type = getattr(jssobjects, child.tag)
            objects = [obj_type(self, obj) for obj in child]
            all_objects[child.tag] = QuerySet(objects)

        return all_objects

    def scrape(self, url_path, session_id=None):
        """This method allows for access to things that don't have an API.

        It's entirely up to you how to implement handling the requests/responses because there's no way Jamf will
        support this.

        Args:
            url_path: String the path to scrape
            session_id: String the session id to use. You can extract a session cookie from a previous response,
                probably using JSESSIONID
        """
        if session_id is None:
            response = self.session.post(self.base_url, data={'username': self.user, 'password': self.password})

            if response.status_code == 200:
                scrape_url = '{}/{}'.format(self.base_url, url_path)
                return self.session.get(scrape_url)

    def version(self):
        return self.JSSUser().version.text

# There's a lot of repetition involved in creating the object query
# methods on JSS, so we create them dynamically at import time.
def add_search_method(cls, name):
    """Add a class-specific search method to a class (JSS)"""
    # Get the actual class to search for, from str `name`
    obj_type = getattr(jssobjects, name)

    # Create a closure over the retrieved class to do our search.
    def api_method(self, data=None, **kwargs):
        """Flexibly search the JSS for objects of type {0}.

            Args:
                data (int, str, :obj:`xml.etree.ElementTree.Element`, optional): Argument to query for.

                    Different queries are performed depending on the type of this arg:
                        - **None** (or provide no argument / default):
                            Search for all objects.
                        - **int**: Search for an object by ID.
                        - **str**: Search for an object by name. Some objects
                            allow 'match' searches, using '*' as the
                            wildcard operator.
                        - :obj:`xml.etree.ElementTree.Element`: create a new
                            object from the Element's data.

                **kwargs:
                    {1}

                    Some classes allow additional filters, subsets, etc,
                    in their queries. Check the object's `allowed_kwargs`
                    attribute for a complete list of implemented keys.

                    Not all classes offer all types of searches, nor are
                    they all necessarily offered in a single query.
                    Consult the Casper API documentation for usage.

                    In general, the key name is applied to the end of the
                    URL, followed by the val; e.g.
                    '<url>/subset/general'.

                    Some common types of extra arguments:

                    subset (list of str or str): XML subelement tags to
                        request (e.g.  ['general', 'purchasing']), OR an
                        '&' delimited string (e.g.
                        'general&purchasing').  Defaults to None.
                    start_date/end_date (str or datetime): Either dates
                        in the form 'YYYY-MM-DD' or a datetime.datetime
                        object.

            Returns:
                QuerySet: If data=None, return all objects of this
                    type.
                {0}: If searching or creating new objects, return an
                    instance of that object.
                None: (FUTURE) Will return None if nothing is found that
                    matches the search criteria.

            Raises:
                GetError for nonexistent objects.
        """
        if not isinstance(data, ElementTree.Element):
            url = obj_type.build_query(data, **kwargs)
            data = self.get(url)

        # TODO: Deprecated and pending removal
        if hasattr(obj_type, "container"):
            data = data.find(obj_type.container)

        if data.find("size") is not None:
            return QuerySet.from_response(obj_type, data, self, **kwargs)
        else:
            return obj_type(self, data)

    # Add in the missing variables to the docstring and set name.
    if hasattr(obj_type, 'allowed_kwargs') and obj_type.allowed_kwargs:
        allowed = ', '.join(obj_type.allowed_kwargs)
        msg = 'Allowed keyword arguments for this class are:\n{}{}'
        kwarg_doc = msg.format(6 * "    ", allowed) if allowed else ""
    else:
        kwarg_doc = "(None supported)"
    api_method.__doc__ = api_method.__doc__.format(name, kwarg_doc)
    api_method.__name__ = name
    # Add the method to the class with the correct name.
    setattr(cls, name, api_method)


# TODO: DRY.. this is just a temp copy/paste to try abstracting
def add_uapi_search_method(cls, name):
    """Add a class-specific search method to a class (JSS)"""
    # Get the actual class to search for, from str `name`
    obj_type = getattr(uapiobjects, name)

    # Create a closure over the retrieved class to do our search.
    def api_method(self, data=None, **kwargs):
        """Flexibly search the JSS for objects of type {0}.

            Args:
                data (None, int, str, xml.etree.ElementTree.Element):
                    Argument to query for. Different queries are
                    performed depending on the type of this arg:
                        None (or provide no argument / default):
                            Search for all objects.
                        int: Search for an object by ID.
                        str: Search for an object by name. Some objects
                            allow 'match' searches, using '*' as the
                            wildcard operator.
                        xml.etree.ElementTree.Element: create a new
                            object from the Element's data.
                kwargs:
                    {1}

                    Some classes allow additional filters, subsets, etc,
                    in their queries. Check the object's `allowed_kwargs`
                    attribute for a complete list of implemented keys.

                    Not all classes offer all types of searches, nor are
                    they all necessarily offered in a single query.
                    Consult the Casper API documentation for usage.

                    In general, the key name is applied to the end of the
                    URL, followed by the val; e.g.
                    '<url>/subset/general'.

                    Some common types of extra arguments:

                    subset (list of str or str): XML subelement tags to
                        request (e.g.  ['general', 'purchasing']), OR an
                        '&' delimited string (e.g.
                        'general&purchasing').  Defaults to None.
                    start_date/end_date (str or datetime): Either dates
                        in the form 'YYYY-MM-DD' or a datetime.datetime
                        object.

            Returns:
                QuerySet: If data=None, return all objects of this
                    type.
                {0}: If searching or creating new objects, return an
                    instance of that object.
                None: (FUTURE) Will return None if nothing is found that
                    matches the search criteria.

            Raises:
                GetError for nonexistent objects.
        """
        if not isinstance(data, dict):
            url = obj_type.build_query(data, **kwargs)
            data = self.jss.get(url, headers={'Content-Type': 'application/json', 'Accept': 'application/json'},
                            auth=UAPIAuth(self.jss.user, self.jss.password, "{}/uapi/auth/tokens".format(self.jss.base_url)))

        if isinstance(data, list):
            return [obj_type(self.jss, d) for d in data]
        else:
            return obj_type(self.jss, data)

    # Add in the missing variables to the docstring and set name.
    if hasattr(obj_type, 'allowed_kwargs') and obj_type.allowed_kwargs:
        allowed = ', '.join(obj_type.allowed_kwargs)
        msg = 'Allowed keyword arguments for this class are:\n{}{}'
        kwarg_doc = msg.format(6 * "    ", allowed) if allowed else ""
    else:
        kwarg_doc = "(None supported)"
    api_method.__doc__ = api_method.__doc__.format(name, kwarg_doc)
    api_method.__name__ = name
    # Add the method to the class with the correct name.
    setattr(cls, name, api_method)


# Run `add_search_method` against everything that jss.jssobjects exports.
for jss_class in jssobjects.__all__:
    add_search_method(JSS, jss_class)

for jss_uapi_class in uapiobjects.__all__:
    add_uapi_search_method(JSS.UAPI, jss_uapi_class)

# pylint: disable=too-many-instance-attributes, too-many-public-methods


class JSSObjectFactory(object):
    """Deprecated"""
    pass



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""distribution_points.py

Utility class for synchronizing packages to Jamf file repositories,
CDPs, and JDSs.
"""


from __future__ import absolute_import
import os

from .distribution_point import (
    AFPDistributionPoint, SMBDistributionPoint, JDS, CDP, LocalRepository, AWS, JCDS)
from .exceptions import JSSError
from .tools import (is_osx, is_linux, is_package)


class DistributionPoints(object):
    """Manage multiple DistributionPoint objects.

    DistributionPoints is an object which reads DistributionPoint
    configuration data from the JSS and serves as a container for
    objects representing the configured distribution points.

    This class provides an abstracted interface for uploading packages
    and dmg's to file repositories.

    PLEASE NOTE: Not all DistributionPoint types support all of the
    available methods, or in the same way. For example, a JDS has
    caveats to the reliability of the exists() method.

    Support for Local repositories, AFP/SMB shares and JDS and CDP
    servers are included, and are selected based on configuration files.

    This object can copy files to multiple repositories, avoiding the
    need to use Casper Admin to "Replicate" from one repository to
    another, as long as the repositories are all configured correctly.

    See the individual Repository subclasses for information regarding
    type-specific properties and configuration.
    """

    def __init__(self, jss):
        """Config the DP dict from configuration file.

        The JSS API endpoint DistributionPoints is used to automatically
        configure AFP and SMB shares. To make use of this, the repo's
        dictionary should contain only the name of the repo, as found in
        the web interface, and the password for the RW user.

        Please see the docstrings for the different DistributionPoint
        subclasses for information regarding required configuration
        information and properties.

        Args:
            jss: JSS server object

        Raises:
            JSSError if an unsupported OS is used.
        """
        self.jss = jss
        self._children = []

        # If no distribution points are configured, there's nothing to
        # do here.
        if self.jss.repo_prefs:
            self.dp_info = self.jss.DistributionPoint()

            for repo in self.jss.repo_prefs:
                # Handle AFP/SMB shares, as they can be auto-configured.
                # Legacy system did not require explicit type key.
                if not repo.get("type"):
                    # Must be AFP or SMB.
                    # Use JSS.DistributionPoints information to
                    # automatically configure this DP.
                    dpt = self._get_auto_configured_dp(repo)
                # Handle Explictly declared DP's.
                elif repo.get("type") in ["AFP", "SMB"]:
                    dpt = self._get_explicitly_configured_dp(repo)
                elif repo.get("type") == "JDS":
                    dpt = JDS(jss=self.jss)
                elif repo.get("type") == "CDP":
                    dpt = CDP(jss=self.jss)
                elif repo.get("type") == "Local":
                    mount_point = repo["mount_point"]
                    share_name = repo["share_name"]
                    dpt = LocalRepository(mount_point=mount_point,
                                          share_name=share_name, jss=self.jss)
                elif repo.get("type") == "JCDS":
                    dpt = JCDS(jss=self.jss)
                elif repo.get("type") == "AWS":
                    dpt = AWS(jss=self.jss, **repo)
                else:
                    raise ValueError("Distribution Point Type not recognized.")

                # Add the DP to the list.
                self._children.append(dpt)

    def __iter__(self):
        for dp in self._children:
            yield dp

    def __len__(self):
        return len(self._children)

    def _get_auto_configured_dp(self, repo):
        "Return a file share DP from auto-configured data."""
        dpt = None
        for dp_object in self.dp_info:
            if repo["name"] == dp_object.name:
                url = dp_object.findtext("ip_address")
                connection_type = dp_object.findtext("connection_type")
                share_name = dp_object.findtext("share_name")
                domain = dp_object.findtext("workgroup_or_domain")
                port = dp_object.findtext("share_port")
                username = dp_object.findtext("read_write_username")
                password = repo.get("password")
                # Make very sure this password is unicode.
                if isinstance(password, str):
                    password = password.encode("utf-8")

                if is_osx():
                    mount_point = os.path.join("/Volumes", share_name)
                elif is_linux():
                    mount_point = os.path.join("/mnt", share_name)
                else:
                    raise JSSError("Unsupported OS.")

                if connection_type == "AFP":
                    dpt = AFPDistributionPoint(
                        url=url, port=port, share_name=share_name,
                        mount_point=mount_point, username=username,
                        password=password, jss=self.jss)
                elif connection_type == "SMB":
                    dpt = SMBDistributionPoint(
                        url=url, port=port, share_name=share_name,
                        mount_point=mount_point, domain=domain,
                        username=username, password=password,
                        jss=self.jss)

        if not dpt:
            raise ValueError(
                "Error auto-configuring distribution point '{}'!".format(
                    repo["name"]))

        return dpt

    def _get_explicitly_configured_dp(self, repo):
        "Return a file share DP from auto-configured data."""
        url = repo["URL"]

        # If found, strip the scheme off the URL
        if "://" in url:
            url = url.split("://")[1]

        connection_type = repo["type"]
        share_name = repo["share_name"]

        # Domain is not used for AFP.
        domain = repo.get("domain")
        username = repo["username"]
        password = repo["password"]
        # Make very sure this password is unicode.
        if isinstance(password, str):
            password = password.encode("utf-8")

        if is_osx():
            mount_point = os.path.join("/Volumes", share_name)
        elif is_linux():
            mount_point = os.path.join("/mnt", share_name)
        else:
            raise JSSError("Unsupported OS.")

        if connection_type == "AFP":

            # If port isn't given, assume it's the std of
            # 548.
            port = repo.get("port", "548")
            dpt = AFPDistributionPoint(
                url=url, port=port, share_name=share_name,
                mount_point=mount_point, username=username, password=password,
                jss=self.jss)
        elif connection_type == "SMB":
            # If port isn't given, assume it's the std of
            # 445.
            port = repo.get("port", "445")
            dpt = SMBDistributionPoint(
                url=url, port=port, share_name=share_name,
                mount_point=mount_point, domain=domain, username=username,
                password=password, jss=self.jss)

        return dpt

    def add_distribution_point(self, distribution_point):
        """Add a distribution point to the list."""
        self._children.append(distribution_point)

    def remove_distribution_point(self, index):
        """Remove a distribution point by index."""
        self._children.pop(index)

    def copy(self, filename, id_=-1, pre_callback=None, post_callback=None):
        """Copy a package to all repos.

        Determines appropriate location (for file shares) and type based
        on file extension.

        Args:
            filename: String path to the local file to copy.
            id_: Package object ID to target. For use with JDS and CDP
                DP's only. If uploading a package that does not have a
                corresponding object, use id_ of -1, which is the
                default.
            pre_callback: Func to call before each distribution point
                starts copying. Should accept a Repository connection
                dictionary as a parameter. Will be called like:
                `pre_callback(repo.connection)`
            post_callback: Func to call after each distribution point
                finishes copying. Should accept a Repository connection
                dictionary as a parameter. Will be called like:
                `pre_callback(repo.connection)`
        """
        for repo in self._children:
            if is_package(filename):
                copy_method = repo.copy_pkg
            if pre_callback:
                pre_callback(repo.connection)
            copy_method(filename, id_)
            if post_callback:
                post_callback(repo.connection)

    def copy_pkg(self, filename, id_=-1):
        """Copy a pkg, dmg, or zip to all repositories.

        Args:
            filename: String path to the local file to copy.
            id_: Integer ID you wish to associate package with for a JDS
                or CDP only. Default is -1, which is used for creating
                a new package object in the database.
        """
        for repo in self._children:
            repo.copy_pkg(filename, id_)

    def delete(self, filename):
        """Delete a file from all repositories which support it.

        Individual repositories will determine correct location to
        delete from (Scripts vs. Packages).

        This will not remove the corresponding Package or Script object
        from the JSS's database!

        Args:
            filename: The filename you wish to delete (do not include a
                path).
        """
        for repo in self._children:
            if hasattr(repo, "delete"):
                repo.delete(filename)

    def mount(self):
        """Mount all mountable distribution points."""
        for child in self._children:
            if hasattr(child, "mount"):
                child.mount()

    def umount(self, forced=True):
        """Umount all mountable distribution points.

        Defaults to using forced method.
        """
        for child in self._children:
            if hasattr(child, "umount"):
                child.umount(forced)

    def exists(self, filename):
        """Report whether a file exists on all distribution points.

        Determines file type by extension.

        Args:
            filename: Filename you wish to check. (No path! e.g.:
                "AdobeFlashPlayer-14.0.0.176.pkg")

        Returns:
            Boolean
        """
        result = True
        for repo in self._children:
            if not repo.exists(filename):
                result = False
        return result

    def __repr__(self):
        """Print out information on distribution points."""
        output = []
        index = 0
        for child in self._children:
            output.append("%s" % (79 * "-"))
            output.append("Index: %s" % index)
            output.append(child.__repr__())
            index += 1

        return "\n".join(output)



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""tools.py

Helper functions for python-jss.
"""


from __future__ import absolute_import
import copy
from functools import wraps
import os
import re
try:
    from urllib import quote  # Python 2.X
except ImportError:
    from urllib.parse import quote  # Python 3+
from xml.etree import ElementTree


PKG_TYPES = {".PKG", ".DMG", ".ZIP"}


def is_osx():
    """Convenience function for testing OS version."""
    result = True if os.uname()[0] == "Darwin" else False
    return result


def is_linux():
    """Convenience function for testing OS version."""
    result = True if os.uname()[0] == "Linux" else False
    return result


def is_package(filename):
    """Return True if filename is a package type.

    Args:
        filename: String filename with no path.
    """
    # TODO: Possible rewording
    # return filename.upper().endswith(tuple(PKG_TYPES))
    return os.path.splitext(filename)[1].upper() in PKG_TYPES


def is_script(filename):
    """Return True if a filename is NOT a package.

    Because there are so many script types, it's easier to see if
    the file is a package than to see if it is a script.

    Args:
        filename: String filename with no path.
    """
    return not is_package(filename)


def convert_response_to_text(response):
    """Convert a JSS HTML response to plaintext."""
    # Responses are sent as html. Split on the newlines and give us
    # the <p> text back.
    errorlines = response.content.decode("utf-8").split("\n")
    error = []
    pattern = re.compile(r"<p.*>(.*)</p>")
    for line in errorlines:
        content_line = re.search(pattern, line)
        if content_line:
            error.append(content_line.group(1))

    return ". ".join(error) + " {}.".format(response.url)


def error_handler(exception_cls, response):
    """Handle HTTP errors by formatting into strings."""
    # Responses are sent as html. Split on the newlines and give us
    # the <p> text back.
    error = convert_response_to_text(response)
    exception = exception_cls("Response Code: %s\tResponse: %s" %
                              (response.status_code, error))
    exception.status_code = response.status_code
    raise exception


def loop_until_valid_response(prompt):
    """Loop over entering input until it is a valid bool-ish response.

    Args:
        prompt: Text presented to user.

    Returns:
        The bool value equivalent of what was entered.
    """
    responses = {"Y": True, "YES": True, "TRUE": True,
                 "N": False, "NO": False, "FALSE": False}
    response = ""
    while response.upper() not in responses:
        response = raw_input(prompt)

    return responses[response.upper()]


def indent_xml(elem, level=0, more_sibs=False):
    """Indent an xml element object to prepare for pretty printing.

    To avoid changing the contents of the original Element, it is
    recommended that a copy is made to send to this function.

    Args:
        elem: Element to indent.
        level: Int indent level (default is 0)
        more_sibs: Bool, whether to anticipate further siblings.
    """
    i = "\n"
    pad = "    "
    if level:
        i += (level - 1) * pad
    num_kids = len(elem)
    if num_kids:
        if not elem.text or not elem.text.strip():
            elem.text = i + pad
            if level:
                elem.text += pad
        count = 0
        for kid in elem:
            if kid.tag == "data":
                kid.text = "*DATA*"
            indent_xml(kid, level + 1, count < num_kids - 1)
            count += 1
        if not elem.tail or not elem.tail.strip():
            elem.tail = i
            if more_sibs:
                elem.tail += pad
    else:
        if level and (not elem.tail or not elem.tail.strip()):
            elem.tail = i
            if more_sibs:
                elem.tail += pad


def element_str(elem):
    """Return a string with indented XML data."""
    # deepcopy so we don't mess with the valid XML.
    pretty_data = copy.deepcopy(elem)
    indent_xml(pretty_data)
    return ElementTree.tostring(pretty_data, encoding='unicode')


def quote_and_encode(string):
    """Encode a bytes string to UTF-8 and then urllib.quote"""
    return quote(string.encode('UTF_8'))


def triggers_cache(func):
    """Decorator for enabling methods to trigger cache filling."""

    @wraps(func)
    def trigger_cache(self, *args, **kwargs):
        if hasattr(self, 'cached') and not self.cached:
            self.retrieve()
        return func(self, *args, **kwargs)

    return trigger_cache


def decorate_class_with_caching(cls, methods):
    for method_name in methods:
        if hasattr(cls, method_name):  # JSSObject does not have copy ?
            decorated_method = triggers_cache(getattr(cls, method_name))
            setattr(cls, method_name, decorated_method)



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jss_prefs.py

Class for using a plist for storing python-jss preferences, with an
interactive configuration that runs if you have not previously set up
a plist.
"""
from __future__ import print_function

from __future__ import absolute_import
import getpass
import os
import readline   # pylint: disable=unused-import
import subprocess
from xml.parsers.expat import ExpatError

from .exceptions import JSSError, GetError
from .jamf_software_server import JSS
from .tools import is_osx, is_linux, loop_until_valid_response
try:
    from .contrib import FoundationPlist as plistlib
except ImportError:
    # If using OSX, FoundationPlist will need Foundation/PyObjC
    # available, or it won't import.
    import plistlib


PREFS_DEFAULT = "com.github.sheagcraig.python-jss.plist"


class JSSPrefs(object):
    """Object representing JSS credentials and configuration.

    This JSSPrefs object can be used as an argument for a new JSS.
    By default and with no arguments, it uses the preference domain
    "com.github.sheagcraig.python-jss.plist". However, alternate
    configurations can be supplied to the __init__ method to use
    something else.

    If no preference file is found, an interactive configuration
    function will try to help configure python-jss.

    The preference file should include the following keys:

        - **jss_url**: String, full path, including port, to JSS, e.g.
          "https://mycasper.donkey.com:8443".
        - **jss_user**: String, API username to use.
        - **jss_pass**: String, API password.
        - **verify**: *(Optional)* Boolean for whether to verify the JSS's
          certificate matches the SSL traffic. This certificate must
          be in your keychain. Defaults to True.
        - **suppress_warnings**: *(Optional)* Boolean for whether to suppress
          the urllib3 warnings likely spamming you if you choose not
          to set verify=False. Enabled by default when verify=False.
        - **repos:** *(Optional)* A list of file repositories dicts to connect.
            - repos dicts:

                CDP and JDS distribution points require:

                - **name:** String name of the distribution point. Must match
                  the value on the JSS.
                - **password:** String password for the read/write user.

                This form uses the distributionpoints API call to determine
                the remaining information. There is also an explicit form;
                See the distribution_points module for more info

                CDP, JDS and AWS types require one dict for the master, with
                key:

                - **type:** String, "CDP", "JDS" or "AWS".

                AWS distribution points require:

                - **aws_access_key_id:** The access key ID for the user with r/w permission
                    to the jamf bucket
                - **aws_secret_access_key:** The secret key for this user. NOTE: to avoid storing sensitive credentials
                    this can also be read from the environment variable `AWS_SECRET_ACCESS_KEY`.

    """

    def __init__(self, preferences_file=None):
        """Create a preferences object.

        This JSSPrefs object can be used as an argument for a new JSS.
        By default and with no arguments, it uses the preference domain
        "com.github.sheagcraig.python-jss.plist". However, alternate
        configurations can be supplied to the __init__ method to use
        something else.

        If no preferences file is specified, an interactive config
        method will run to help set up python-jss.

        See the JSSPrefs __doc__ for information on supported
        preferences.

        Args:
            preferences_file: String path to an alternate location to
            look for preferences. Defaults based on OS are:
                OS X: "~/Library/Preferences/com.github.sheagcraig.python-jss.plist"
                Linux: "~/.com.github.sheagcraig.python-jss.plist"

        Raises:
            JSSError if using an unsupported OS.
        """
        if preferences_file is None:
            if is_osx():
                preferences_file = os.path.join(
                    "~", "Library", "Preferences", PREFS_DEFAULT)
            elif is_linux():
                preferences_file = os.path.join("~", "." + PREFS_DEFAULT)
            else:
                raise JSSError("Unsupported OS.")

        self.preferences_file = os.path.expanduser(preferences_file)

        if not os.path.exists(self.preferences_file):
            self.configure()

        self._parse_plist()

    def _parse_plist(self):
        """Try to reset preferences from preference_file."""
        # If there's an ExpatError, it's probably because the plist is
        # in binary plist format.
        try:
            prefs = plistlib.readPlist(self.preferences_file)
        except ExpatError:
            # If we're on OSX, try to convert using another tool.
            if is_osx():
                plist = subprocess.check_output(
                    ["plutil", "-convert", "xml1", "-o", "-",
                     preferences_file]).decode()
                prefs = plistlib.readPlistFromString(preferences_file)

        self.user = prefs.get("jss_user")
        self.password = prefs.get("jss_pass")
        self.url = prefs.get("jss_url")
        if not all([self.user, self.password, self.url]):
            raise TypeError("Please provide all required preferences!")

        # Optional file repository array. Defaults to empty list.
        self.repos = []
        for repo in prefs.get("repos", []):
            self.repos.append(dict(repo))

        self.verify = prefs.get("verify", True)
        self.suppress_warnings = prefs.get("suppress_warnings", True)

    def configure(self):
        """Prompt user for config and write to plist

        Uses preferences_file argument from JSSPrefs.__init__ as path
        to write.
        """
        prefs = {}
        print ("It seems like you do not have a preferences file configured. "
               "Please answer the following questions to generate a plist at "
               "%s for use with python-jss." % self.preferences_file)

        prefs["jss_url"] = raw_input(
            "The complete URL to your JSS, with port (e.g. "
            "'https://mycasperserver.org:8443')\nURL: ")

        prefs["jss_user"] = raw_input("API Username: ")
        prefs["jss_pass"] = getpass.getpass("API User's Password: ")
        verify_prompt = ("Do you want to verify that traffic is encrypted by "
                         "a certificate that you trust?: (Y|N) ")
        prefs["verify"] = loop_until_valid_response(verify_prompt)
        prefs["repos"] = self._handle_repos(prefs)

        plistlib.writePlist(prefs, self.preferences_file)
        print("Preferences created.\n")

    def _handle_repos(self, prefs):
        """Handle repo configuration."""
        repos_array = []

        # Make a temporary jss object to try to pull repo information.
        jss_server = JSS(
            url=prefs["jss_url"], user=prefs["jss_user"],
            password=prefs["jss_pass"], ssl_verify=prefs["verify"])
        print("Fetching distribution point info...")
        try:
            dpts = jss_server.DistributionPoint()
        except GetError:
            print (
                "Fetching distribution point info failed. If you want to "
                "configure distribution points, ensure that your API user "
                "has read permissions for distribution points, and that the "
                "URL, username, and password are correct.")
            dpts = None

        if dpts:
            print ("There are file share distribution points configured on "
                   "your JSS. Most of the configuration can be automated "
                   "from the information on the JSS, with the exception of "
                   "the password for the R/W user.\n")

            for dpt in dpts:
                repo_dict = {}
                repo_dict["name"] = dpt.get("name")
                repo_pass_string = getpass.getpass(
                    "Please enter the R/W user's password for distribution "
                    "point: %s: " % dpt.get("name", "<NO NAME CONFIGURED>"))
                repo_dict["password"] = repo_pass_string
                repos_array.append(repo_dict)

        jds = _handle_dist_server("JDS")
        if jds:
            repos_array.append(jds)
        cdp = _handle_dist_server("CDP")
        if cdp:
            repos_array.append(cdp)

        return repos_array


def _handle_dist_server(ds_type):
    """Ask user for whether to use a type of dist server."""
    if ds_type not in ("JDS", "CDP"):
        raise ValueError("Must be JDS or CDP")
    prompt = "Does your JSS use a %s? (Y|N): " % ds_type
    result = loop_until_valid_response(prompt)

    if result:
        return {"type": ds_type}



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""curl_adapter.py

Adapter object to provide the Requests API wrapped around curl.

macOS ships with a combination of python and openssl that cannot do TLS,
which is required to work with current JSS versions.

python-jss offers two options:
1. Add the Requests library, along with updating a number of its
   dependencies.
2. Use subprocess to funnel web requests through curl, which is built on
   Macs against the Cocoa networking frameworks.

This module provides the second option, a wrapper around Curl that uses
the same API as Requests, to facilitate replacing the networking layer
for more advanced users. CurlAdapter is the default when instantiating a
JSS object beginning with python-jss 2.0.0.
"""


from __future__ import absolute_import
import copy
import subprocess
import logging
import sys

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
logger.addHandler(logging.StreamHandler())

from .exceptions import JSSError, SSLVerifyError

CURL_RETURNCODE = {
    1: 'Unsupported protocol. This build of curl has no support for this protocol.',
    2: 'Failed to initialize.',
    3: 'URL malformed. The syntax was not correct.',
    4: 'A feature or option that was needed to perform the desired request was not enabled or was explicitly disabled '
       'at build-time. To make curl able to do this, you probably need another build of libcurl!',
    5: 'Couldn\'t resolve proxy. The given proxy host could not be resolved.',
    6: 'Couldn\'t resolve host. The given remote host was not resolved.',
    7: 'Failed to connect to host.',
    8: 'Weird server reply. The server sent data curl couldn\'t parse.',
    22: 'HTTP page not retrieved. The requested url was not found or returned another error with the HTTP error '
        'code being 400 or above.',
    23: 'Write error. Curl couldn\'t write data to a local filesystem or similar.',
    27: 'Out of memory. A memory allocation request failed.',
    28: 'Operation timeout. The specified time-out period was reached according to the conditions.',
    33: 'HTTP range error. The range "command" didn\'t work.',
    35: 'SSL connect error. The SSL handshaking failed.',
    47: 'Too many redirects. When following redirects, curl hit the maximum amount.',
    60: 'Peer certificate cannot be authenticated with known CA certificates.'
}

# Map Python 2 unicode type for Python 3.
if sys.version_info.major == 3:
    unicode = str

class CurlAdapter(object):
    """Adapter to use Curl for all Casper API calls

    Attributes:
        auth (2-tuple of str): Username and password for making requests.
        ssl_verify (bool): Whether to verify SSL traffic. Defaults to
            True.
        use_tls: Whether to use TLS. Defaults to True.
    """
    base_headers = ['Accept: application/xml']

    def __init__(self, verify=True):
        self.auth = ('', '')
        self.verify = verify
        self.use_tls = True

    def get(self, url, headers=None):
        return self._request(url, headers)

    def post(self, url, data=None, headers=None, files=None):
        content_type = 'text/xml' if not files else 'multipart/form-data'
        header = ['Content-Type: {}'.format(content_type)]
        if headers:
            [header.append('{}: {}'.format(k, headers[k])) for k in headers]

        post_kwargs = {"--request": "POST"}
        return self._request(url, header, data, files, **post_kwargs)

    def put(self, url, data=None, headers=None, files=None):
        content_type = 'text/xml' if not files else 'multipart/form-data'
        header = ['Content-Type: {}'.format(content_type)]
        if headers:
            [header.append('{}: {}'.format(k, headers[k])) for k in headers]

        put_args = {"--request": "PUT"}
        return self._request(url, header, data, files, **put_args)

    def delete(self, url, data=None, headers=None):
        delete_args = {"--request": "DELETE"}
        if data:
            headers += ['Content-Type: text/xml']
            delete_args['--data'] = data
        return self._request(url, headers, **delete_args)

    def _request(self, url, headers=None, data=None, files=None, **kwargs):
        command = self._build_command(url, headers, data, files, **kwargs)

        # Ensure all arguments to curl are encoded. This is the last
        # point of contact, so just do it here and keep it Unicode
        # everywhere else.
        command = [
            item.encode('UTF-8') if isinstance(item, unicode) else item
            for item in command]

        logger.debug(' '.join(command))

        try:
            response = subprocess.check_output(command).decode()
        except subprocess.CalledProcessError as err:
            if err.returncode in CURL_RETURNCODE:
                raise JSSError('CURL Error: {}'.format(CURL_RETURNCODE[err.returncode]))
            else:
                raise JSSError('Unknown curl error: {}'.format(err.returncode))

        return CurlResponseAdapter(response, url)

    def _build_command(
        self, url, headers=None, data=None, files=None, **kwargs):
        """Construct the argument list for curl.

        Encode all unicode to bytes with UTF-8 on the way out.

        Args:
            url (str): Full URL to request.
            headers (sequence of str): Header strings to use. Defaults to
                None.
            kwargs (str: str): Extra commandline arguments and their
                values to be added to the curl command.

        Returns:
            list of arguments to subprocess for making request via curl.
        """
        # Curl expects auth information as a ':' delimited string.
        auth = '{}:{}'.format(*self.auth)
        command = ["curl", "-u", auth]

        # Remove the progress bar that curl displays in a subprocess.
        command.append("--silent")

        # Add the returncode to the output so we can parse it into
        # the resulting CurlResponseAdapter.
        command += ["--write-out", "|%{response_code}"]

        if self.verify == False:
            command.append("--insecure")

        if self.use_tls:
            command.append("--tlsv1")

        compiled_headers = copy.copy(self.base_headers)
        if headers:
            compiled_headers += headers
        for header in compiled_headers:
            command += ['--header', header]

        if data:
            if isinstance(data, file):
                command += ["--data-binary", "@{}".format(data.name)]
            elif isinstance(data, dict):
                [command.extend(["-F", "{}={}".format(k, data[k])]) for k in data]
            else:
                command += ["--data", data]

        if files:
            path = files['name'][1].name
            content_type = files['name'][2]
            file_data = 'name=@{};type={}'.format(path, content_type)
            command += ["--form", file_data]

        for key, val in kwargs.items():
            command += [key, val]

        command.append(url)

        return command

    def suppress_warnings(self):
        """Included for compatibility with RequestsAdapter"""
        # TODO: Remove
        pass


class CurlResponseAdapter(object):
    """Wrapper for Curl responses"""

    def __init__(self, response, url):
        self.response = response
        self.url = url
        content, _, status_code = response.rpartition("|")
        try:
            self.status_code = int(status_code)
        except ValueError:
            self.status_code = 0
        self.content = content
        # Requests' text attribute returns unicode, so convert curl's
        # returned bytes.
        self.text = content.decode('UTF-8')



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from __future__ import print_function
import os
from setuptools import setup, find_packages

from jss import __version__

# http://stackoverflow.com/questions/10718767/have-the-same-readme-both-in-markdown-and-restructuredtext
try:
    from pypandoc import convert

    read_md = lambda f: convert(f, 'rst')
except ImportError:
    print("Warning: pypandoc module not found, could not convert md to rst")
    read_md = lambda f: open(os.path.join(os.path.dirname(__file__), f), 'r').read()

setup(name='python-jss',
      version=__version__,
      packages=find_packages(),
      description='Python wrapper for JSS API.',
      long_description=read_md('README.md'),
      author='Shea G. Craig',
      url='https://github.com/JSSImporter/python-jss/',
      license='GPLv3',
      install_requires=['requests>=2.24.0'],
      extras_require={
          'reST': [
              "Sphinx>=3.1.2", "docutils>=0.16"]
      },
      setup_requires=['pytest-runner'],
      tests_require=[
          'pytest', 'coverage', 'pytest-docker'
      ],
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: GNU General Public License v3 (GPLv3)']
      )



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""exceptions.py

Custom Exceptions for python-jss.
"""


class JSSError(Exception):
    """Base python-jss exception class."""
    pass


class RequestError(JSSError):

    def __init__(self, *args, **kwargs):
        super(RequestError, self).__init__(*args, **kwargs)
        self.status_code = None


class GetError(RequestError):
    """GET exception."""
    pass


class PutError(RequestError):
    """PUT exception."""
    pass


class PostError(RequestError):
    """POST exception."""
    pass


class DeleteError(RequestError):
    """DEL exception."""
    pass


class MethodNotAllowedError(JSSError):
    """JSSObject is not allowed to use HTTP method."""
    pass


class SSLVerifyError(JSSError):
    """Server's certificate could not be verified."""
    pass



# encoding: utf-8
#
# Copyright 2009-2020 Greg Neagle.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
gurl.py

Created by Greg Neagle on 2013-11-21.
Modified in Feb 2016 to add support for NSURLSession.
Updated June 2019 for compatibility with Python 3 and PyObjC 5.1.2+

curl replacement using NSURLConnection and friends

Tested with PyObjC 2.5.1 (inlcuded with macOS)
and with PyObjC 5.2b1. Should also work with PyObjC 5.1.2.
May fail with other versions of PyObjC due to issues with completion handler
signatures.
"""
from __future__ import absolute_import, print_function

import os
import xattr

try:
    # Python 2
    from urlparse import urlparse
except ImportError:
    # Python 3
    from urllib.parse import urlparse


# builtin super doesn't work with Cocoa classes in recent PyObjC releases.
# pylint: disable=redefined-builtin,no-name-in-module
from objc import super
# pylint: enable=redefined-builtin,no-name-in-module

# PyLint cannot properly find names inside Cocoa libraries, so issues bogus
# No name 'Foo' in module 'Bar' warnings. Disable them.
# pylint: disable=E0611


from Foundation import (NSBundle, NSRunLoop, NSData, NSDate,
                        NSObject, NSURL, NSURLConnection,
                        NSMutableURLRequest,
                        NSURLRequestReloadIgnoringLocalCacheData,
                        NSURLResponseUnknownLength,
                        NSLog,
                        NSURLCredential, NSURLCredentialPersistenceNone,
                        NSPropertyListSerialization,
                        NSPropertyListMutableContainersAndLeaves,
                        NSPropertyListXMLFormat_v1_0)

try:
    from Foundation import NSURLSession, NSURLSessionConfiguration
    from CFNetwork import (kCFNetworkProxiesHTTPSEnable,
                           kCFNetworkProxiesHTTPEnable)
    NSURLSESSION_AVAILABLE = True
except ImportError:
    NSURLSESSION_AVAILABLE = False

# Disable PyLint complaining about 'invalid' names
# pylint: disable=C0103

if NSURLSESSION_AVAILABLE:
    # NSURLSessionAuthChallengeDisposition enum constants
    NSURLSessionAuthChallengeUseCredential = 0
    NSURLSessionAuthChallengePerformDefaultHandling = 1
    NSURLSessionAuthChallengeCancelAuthenticationChallenge = 2
    NSURLSessionAuthChallengeRejectProtectionSpace = 3

    # NSURLSessionResponseDisposition enum constants
    NSURLSessionResponseCancel = 0
    NSURLSessionResponseAllow = 1
    NSURLSessionResponseBecomeDownload = 2

    # TLS/SSLProtocol enum constants
    kSSLProtocolUnknown = 0
    kSSLProtocol3 = 2
    kTLSProtocol1 = 4
    kTLSProtocol11 = 7
    kTLSProtocol12 = 8
    kDTLSProtocol1 = 9

    # define a helper function for block callbacks
    import ctypes
    import objc
    CALLBACK_HELPER_AVAILABLE = True
    try:
        _objc_so = ctypes.cdll.LoadLibrary(
            os.path.join(objc.__path__[0], '_objc.so'))
    except OSError:
        # could not load _objc.so
        CALLBACK_HELPER_AVAILABLE = False
    else:
        PyObjCMethodSignature_WithMetaData = (
            _objc_so.PyObjCMethodSignature_WithMetaData)
        PyObjCMethodSignature_WithMetaData.restype = ctypes.py_object

        def objc_method_signature(signature_str):
            '''Return a PyObjCMethodSignature given a call signature in string
            format'''
            return PyObjCMethodSignature_WithMetaData(
                ctypes.create_string_buffer(signature_str), None, False)

# pylint: enable=E0611

# disturbing hack warning!
# this works around an issue with App Transport Security on 10.11
bundle = NSBundle.mainBundle()
info = bundle.localizedInfoDictionary() or bundle.infoDictionary()
info['NSAppTransportSecurity'] = {'NSAllowsArbitraryLoads': True}


def NSLogWrapper(message):
    '''A wrapper function for NSLog to prevent format string errors'''
    NSLog('%@', message)


ssl_error_codes = {
    -9800: u'SSL protocol error',
    -9801: u'Cipher Suite negotiation failure',
    -9802: u'Fatal alert',
    -9803: u'I/O would block (not fatal)',
    -9804: u'Attempt to restore an unknown session',
    -9805: u'Connection closed gracefully',
    -9806: u'Connection closed via error',
    -9807: u'Invalid certificate chain',
    -9808: u'Bad certificate format',
    -9809: u'Underlying cryptographic error',
    -9810: u'Internal error',
    -9811: u'Module attach failure',
    -9812: u'Valid cert chain, untrusted root',
    -9813: u'Cert chain not verified by root',
    -9814: u'Chain had an expired cert',
    -9815: u'Chain had a cert not yet valid',
    -9816: u'Server closed session with no notification',
    -9817: u'Insufficient buffer provided',
    -9818: u'Bad SSLCipherSuite',
    -9819: u'Unexpected message received',
    -9820: u'Bad MAC',
    -9821: u'Decryption failed',
    -9822: u'Record overflow',
    -9823: u'Decompression failure',
    -9824: u'Handshake failure',
    -9825: u'Misc. bad certificate',
    -9826: u'Bad unsupported cert format',
    -9827: u'Certificate revoked',
    -9828: u'Certificate expired',
    -9829: u'Unknown certificate',
    -9830: u'Illegal parameter',
    -9831: u'Unknown Cert Authority',
    -9832: u'Access denied',
    -9833: u'Decoding error',
    -9834: u'Decryption error',
    -9835: u'Export restriction',
    -9836: u'Bad protocol version',
    -9837: u'Insufficient security',
    -9838: u'Internal error',
    -9839: u'User canceled',
    -9840: u'No renegotiation allowed',
    -9841: u'Peer cert is valid, or was ignored if verification disabled',
    -9842: u'Server has requested a client cert',
    -9843: u'Peer host name mismatch',
    -9844: u'Peer dropped connection before responding',
    -9845: u'Decryption failure',
    -9846: u'Bad MAC',
    -9847: u'Record overflow',
    -9848: u'Configuration error',
    -9849: u'Unexpected (skipped) record in DTLS'}


class Gurl(NSObject):
    '''A class for getting content from a URL
       using NSURLConnection/NSURLSession and friends'''

    # since we inherit from NSObject, PyLint issues a few bogus warnings
    # pylint: disable=W0232,E1002

    # Don't want to define the attributes twice that are initialized in
    # initWithOptions_(), so:
    # pylint: disable=E1101,W0201

    GURL_XATTR = 'com.googlecode.munki.downloadData'

    def initWithOptions_(self, options):
        '''Set up our Gurl object'''
        self = super(Gurl, self).init()
        if not self:
            return None

        self.follow_redirects = options.get('follow_redirects', False)
        self.ignore_system_proxy = options.get('ignore_system_proxy', False)
        self.destination_path = options.get('file')
        self.can_resume = options.get('can_resume', False)
        self.url = options.get('url')
        self.additional_headers = options.get('additional_headers', {})
        self.username = options.get('username')
        self.password = options.get('password')
        self.download_only_if_changed = options.get(
            'download_only_if_changed', False)
        self.cache_data = options.get('cache_data')
        self.connection_timeout = options.get('connection_timeout', 60)
        if NSURLSESSION_AVAILABLE:
            self.minimum_tls_protocol = options.get(
                'minimum_tls_protocol', kTLSProtocol1)

        self.log = options.get('logging_function', NSLogWrapper)

        self.resume = False
        self.response = None
        self.headers = None
        self.status = None
        self.error = None
        self.SSLerror = None
        self.done = False
        self.redirection = []
        self.destination = None
        self.bytesReceived = 0
        self.expectedLength = -1
        self.percentComplete = 0
        self.connection = None
        self.session = None
        self.task = None
        return self

    def start(self):
        '''Start the connection'''
        if not self.destination_path:
            self.log('No output file specified.')
            self.done = True
            return
        url = NSURL.URLWithString_(self.url)
        request = (
            NSMutableURLRequest.requestWithURL_cachePolicy_timeoutInterval_(
                url, NSURLRequestReloadIgnoringLocalCacheData,
                self.connection_timeout))
        if self.additional_headers:
            for header, value in self.additional_headers.items():
                request.setValue_forHTTPHeaderField_(value, header)
        # does the file already exist? See if we can resume a partial download
        if os.path.isfile(self.destination_path):
            stored_data = self.getStoredHeaders()
            if (self.can_resume and 'expected-length' in stored_data and
                    ('last-modified' in stored_data or 'etag' in stored_data)):
                # we have a partial file and we're allowed to resume
                self.resume = True
                local_filesize = os.path.getsize(self.destination_path)
                byte_range = 'bytes=%s-' % local_filesize
                request.setValue_forHTTPHeaderField_(byte_range, 'Range')
        if self.download_only_if_changed and not self.resume:
            stored_data = self.cache_data or self.getStoredHeaders()
            if 'last-modified' in stored_data:
                request.setValue_forHTTPHeaderField_(
                    stored_data['last-modified'], 'if-modified-since')
            if 'etag' in stored_data:
                request.setValue_forHTTPHeaderField_(
                    stored_data['etag'], 'if-none-match')
        if NSURLSESSION_AVAILABLE:
            configuration = (
                NSURLSessionConfiguration.defaultSessionConfiguration())

            # optional: ignore system http/https proxies (10.9+ only)
            if self.ignore_system_proxy is True:
                configuration.setConnectionProxyDictionary_(
                    {kCFNetworkProxiesHTTPEnable: False,
                     kCFNetworkProxiesHTTPSEnable: False})

            # set minimum supported TLS protocol (defaults to TLS1)
            configuration.setTLSMinimumSupportedProtocol_(
                self.minimum_tls_protocol)

            self.session = (
                NSURLSession.sessionWithConfiguration_delegate_delegateQueue_(
                    configuration, self, None))
            self.task = self.session.dataTaskWithRequest_(request)
            self.task.resume()
        else:
            self.connection = NSURLConnection.alloc().initWithRequest_delegate_(
                request, self)

    def cancel(self):
        '''Cancel the connection'''
        if self.connection:
            if NSURLSESSION_AVAILABLE:
                self.session.invalidateAndCancel()
            else:
                self.connection.cancel()
            self.done = True

    def isDone(self):
        '''Check if the connection request is complete. As a side effect,
        allow the delegates to work by letting the run loop run for a bit'''
        if self.done:
            return self.done
        # let the delegates do their thing
        NSRunLoop.currentRunLoop().runUntilDate_(
            NSDate.dateWithTimeIntervalSinceNow_(.1))
        return self.done

    def getStoredHeaders(self):
        '''Returns any stored headers for self.destination_path'''
        # try to read stored headers
        try:
            stored_plist_bytestr = xattr.getxattr(
                self.destination_path, self.GURL_XATTR)
        except (KeyError, IOError):
            return {}
        data = NSData.dataWithBytes_length_(
            stored_plist_bytestr, len(stored_plist_bytestr))
        dataObject, _plistFormat, error = (
            NSPropertyListSerialization.
            propertyListFromData_mutabilityOption_format_errorDescription_(
                data, NSPropertyListMutableContainersAndLeaves, None, None))
        if error:
            return {}
        return dataObject

    def storeHeaders_(self, headers):
        '''Store dictionary data as an xattr for self.destination_path'''
        plistData, error = (
            NSPropertyListSerialization.
            dataFromPropertyList_format_errorDescription_(
                headers, NSPropertyListXMLFormat_v1_0, None))
        if error:
            byte_string = b''
        else:
            try:
                byte_string = bytes(plistData)
            except NameError:
                byte_string = str(plistData)
        try:
            xattr.setxattr(self.destination_path, self.GURL_XATTR, byte_string)
        except IOError as err:
            self.log('Could not store metadata to %s: %s'
                     % (self.destination_path, err))

    def normalizeHeaderDict_(self, a_dict):
        '''Since HTTP header names are not case-sensitive, we normalize a
        dictionary of HTTP headers by converting all the key names to
        lower case'''

        # yes, we don't use 'self'!
        # pylint: disable=R0201

        new_dict = {}
        for key, value in a_dict.items():
            new_dict[key.lower()] = value
        return new_dict

    def recordError_(self, error):
        '''Record any error info from completed connection/session'''
        self.error = error
        # If this was an SSL error, try to extract the SSL error code.
        if 'NSUnderlyingError' in error.userInfo():
            ssl_code = error.userInfo()['NSUnderlyingError'].userInfo().get(
                '_kCFNetworkCFStreamSSLErrorOriginalValue', None)
            if ssl_code:
                self.SSLerror = (ssl_code, ssl_error_codes.get(
                    ssl_code, 'Unknown SSL error'))

    def removeExpectedSizeFromStoredHeaders(self):
        '''If a successful transfer, clear the expected size so we
        don\'t attempt to resume the download next time'''
        if str(self.status).startswith('2'):
            # remove the expected-size from the stored headers
            headers = self.getStoredHeaders()
            if 'expected-length' in headers:
                del headers['expected-length']
                self.storeHeaders_(headers)

    def URLSession_task_didCompleteWithError_(self, _session, _task, error):
        '''NSURLSessionTaskDelegate method.'''
        if self.destination and self.destination_path:
            self.destination.close()
            self.removeExpectedSizeFromStoredHeaders()
        if error:
            self.recordError_(error)
        self.done = True

    def connection_didFailWithError_(self, _connection, error):
        '''NSURLConnectionDelegate method
        Sent when a connection fails to load its request successfully.'''
        self.recordError_(error)
        self.done = True
        if self.destination and self.destination_path:
            self.destination.close()

    def connectionDidFinishLoading_(self, _connection):
        '''NSURLConnectionDataDelegate method
        Sent when a connection has finished loading successfully.'''
        self.done = True
        if self.destination and self.destination_path:
            self.destination.close()
            self.removeExpectedSizeFromStoredHeaders()

    def handleResponse_withCompletionHandler_(
            self, response, completionHandler):
        '''Handle the response to the connection'''
        self.response = response
        self.bytesReceived = 0
        self.percentComplete = -1
        self.expectedLength = response.expectedContentLength()

        download_data = {}
        if response.className() == u'NSHTTPURLResponse':
            # Headers and status code only available for HTTP/S transfers
            self.status = response.statusCode()
            self.headers = dict(response.allHeaderFields())
            normalized_headers = self.normalizeHeaderDict_(self.headers)
            if 'last-modified' in normalized_headers:
                download_data['last-modified'] = normalized_headers[
                    'last-modified']
            if 'etag' in normalized_headers:
                download_data['etag'] = normalized_headers['etag']
            download_data['expected-length'] = self.expectedLength

        # self.destination is defined in initWithOptions_
        # pylint: disable=E0203

        if not self.destination and self.destination_path:
            if self.status == 206 and self.resume:
                # 206 is Partial Content response
                stored_data = self.getStoredHeaders()
                if (not stored_data or
                        stored_data.get('etag') != download_data.get('etag') or
                        stored_data.get('last-modified') != download_data.get(
                            'last-modified')):
                    # file on server is different than the one
                    # we have a partial for
                    self.log(
                        'Can\'t resume download; file on server has changed.')
                    if completionHandler:
                        # tell the session task to cancel
                        completionHandler(NSURLSessionResponseCancel)
                    else:
                        # cancel the connection
                        self.connection.cancel()
                    self.log('Removing %s' % self.destination_path)
                    os.unlink(self.destination_path)
                    # restart and attempt to download the entire file
                    self.log(
                        'Restarting download of %s' % self.destination_path)
                    os.unlink(self.destination_path)
                    self.start()
                    return
                # try to resume
                self.log('Resuming download for %s' % self.destination_path)
                # add existing file size to bytesReceived so far
                local_filesize = os.path.getsize(self.destination_path)
                self.bytesReceived = local_filesize
                self.expectedLength += local_filesize
                # open file for append
                self.destination = open(self.destination_path, 'ab')

            elif str(self.status).startswith('2'):
                # not resuming, just open the file for writing
                self.destination = open(self.destination_path, 'wb')
                # store some headers with the file for use if we need to resume
                # the download and for future checking if the file on the server
                # has changed
                self.storeHeaders_(download_data)

        if completionHandler:
            # tell the session task to continue
            completionHandler(NSURLSessionResponseAllow)

    def URLSession_dataTask_didReceiveResponse_completionHandler_(
            self, _session, _task, response, completionHandler):
        '''NSURLSessionDataDelegate method'''
        if CALLBACK_HELPER_AVAILABLE:
            completionHandler.__block_signature__ = objc_method_signature(b'v@i')
        self.handleResponse_withCompletionHandler_(response, completionHandler)

    def connection_didReceiveResponse_(self, _connection, response):
        '''NSURLConnectionDataDelegate delegate method
        Sent when the connection has received sufficient data to construct the
        URL response for its request.'''
        self.handleResponse_withCompletionHandler_(response, None)

    def handleRedirect_newRequest_withCompletionHandler_(
            self, response, request, completionHandler):
        '''Handle the redirect request'''
        def allowRedirect():
            '''Allow the redirect'''
            if completionHandler:
                completionHandler(request)
                return None
            return request

        def denyRedirect():
            '''Deny the redirect'''
            if completionHandler:
                completionHandler(None)
            return None

        newURL = request.URL().absoluteString()
        if response is None:
            # the request has changed the NSURLRequest in order to standardize
            # its format, for example, changing a request for
            # http://www.apple.com to http://www.apple.com/. This occurs because
            # the standardized, or canonical, version of the request is used for
            # cache management. Pass the request back as-is
            # (it appears that at some point Apple also defined a redirect like
            # http://developer.apple.com to https://developer.apple.com to be
            # 'merely' a change in the canonical URL.)
            # Further -- it appears that this delegate method isn't called at
            # all in this scenario, unlike NSConnectionDelegate method
            # connection:willSendRequest:redirectResponse:
            # we'll leave this here anyway in case we're wrong about that
            self.log('Allowing redirect to: %s' % newURL)
            return allowRedirect()
        # If we get here, it appears to be a real redirect attempt
        # Annoyingly, we apparently can't get access to the headers from the
        # site that told us to redirect. All we know is that we were told
        # to redirect and where the new location is.
        self.redirection.append([newURL, dict(response.allHeaderFields())])
        newParsedURL = urlparse(newURL)
        # This code was largely based on the work of Andreas Fuchs
        # (https://github.com/munki/munki/pull/465)
        if self.follow_redirects is True or self.follow_redirects == 'all':
            # Allow the redirect
            self.log('Allowing redirect to: %s' % newURL)
            return allowRedirect()
        elif (self.follow_redirects == 'https'
              and newParsedURL.scheme == 'https'):
            # Once again, allow the redirect
            self.log('Allowing redirect to: %s' % newURL)
            return allowRedirect()
        # If we're down here either the preference was set to 'none',
        # the url we're forwarding on to isn't https or follow_redirects
        # was explicitly set to False
        self.log('Denying redirect to: %s' % newURL)
        return denyRedirect()

    # we don't control the API, so
    # pylint: disable=too-many-arguments
    def URLSession_task_willPerformHTTPRedirection_newRequest_completionHandler_(
            self, _session, _task, response, request, completionHandler):
        '''NSURLSessionTaskDelegate method'''
        self.log(
            'URLSession_task_willPerformHTTPRedirection_newRequest_'
            'completionHandler_')
        if CALLBACK_HELPER_AVAILABLE:
            completionHandler.__block_signature__ = objc_method_signature(b'v@@')
        self.handleRedirect_newRequest_withCompletionHandler_(
            response, request, completionHandler)
    # pylint: enable=too-many-arguments

    def connection_willSendRequest_redirectResponse_(
            self, _connection, request, response):
        '''NSURLConnectionDataDelegate method
        Sent when the connection determines that it must change URLs in order
        to continue loading a request.'''
        self.log('connection_willSendRequest_redirectResponse_')
        return self.handleRedirect_newRequest_withCompletionHandler_(
            response, request, None)

    def connection_canAuthenticateAgainstProtectionSpace_(
            self, _connection, protectionSpace):
        '''NSURLConnection delegate method
        Sent to determine whether the delegate is able to respond to a
        protection space’s form of authentication.
        Deprecated in 10.10'''
        # this is not called in 10.5.x.
        self.log('connection_canAuthenticateAgainstProtectionSpace_')
        if protectionSpace:
            host = protectionSpace.host()
            realm = protectionSpace.realm()
            authenticationMethod = protectionSpace.authenticationMethod()
            self.log('Protection space found. Host: %s Realm: %s AuthMethod: %s'
                     % (host, realm, authenticationMethod))
            if self.username and self.password and authenticationMethod in [
                    'NSURLAuthenticationMethodDefault',
                    'NSURLAuthenticationMethodHTTPBasic',
                    'NSURLAuthenticationMethodHTTPDigest']:
                # we know how to handle this
                self.log('Can handle this authentication request')
                return True
        # we don't know how to handle this; let the OS try
        self.log('Allowing OS to handle authentication request')
        return False

    def handleChallenge_withCompletionHandler_(
            self, challenge, completionHandler):
        '''Handle an authentication challenge'''
        protectionSpace = challenge.protectionSpace()
        host = protectionSpace.host()
        realm = protectionSpace.realm()
        authenticationMethod = protectionSpace.authenticationMethod()
        self.log(
            'Authentication challenge for Host: %s Realm: %s AuthMethod: %s'
            % (host, realm, authenticationMethod))
        if challenge.previousFailureCount() > 0:
            # we have the wrong credentials. just fail
            self.log('Previous authentication attempt failed.')
            if completionHandler:
                completionHandler(
                    NSURLSessionAuthChallengeCancelAuthenticationChallenge,
                    None)
            else:
                challenge.sender().cancelAuthenticationChallenge_(challenge)
        if self.username and self.password and authenticationMethod in [
                'NSURLAuthenticationMethodDefault',
                'NSURLAuthenticationMethodHTTPBasic',
                'NSURLAuthenticationMethodHTTPDigest']:
            self.log('Will attempt to authenticate.')
            self.log('Username: %s Password: %s'
                     % (self.username, ('*' * len(self.password or ''))))
            credential = (
                NSURLCredential.credentialWithUser_password_persistence_(
                    self.username, self.password,
                    NSURLCredentialPersistenceNone))
            if completionHandler:
                completionHandler(
                    NSURLSessionAuthChallengeUseCredential, credential)
            else:
                challenge.sender().useCredential_forAuthenticationChallenge_(
                    credential, challenge)
        else:
            # fall back to system-provided default behavior
            self.log('Allowing OS to handle authentication request')
            if completionHandler:
                completionHandler(
                    NSURLSessionAuthChallengePerformDefaultHandling, None)
            else:
                if (challenge.sender().respondsToSelector_(
                        'performDefaultHandlingForAuthenticationChallenge:')):
                    self.log('Allowing OS to handle authentication request')
                    challenge.sender(
                        ).performDefaultHandlingForAuthenticationChallenge_(
                            challenge)
                else:
                    # Mac OS X 10.6 doesn't support
                    # performDefaultHandlingForAuthenticationChallenge:
                    self.log('Continuing without credential.')
                    challenge.sender(
                        ).continueWithoutCredentialForAuthenticationChallenge_(
                            challenge)

    def connection_willSendRequestForAuthenticationChallenge_(
            self, _connection, challenge):
        '''NSURLConnection delegate method
        Tells the delegate that the connection will send a request for an
        authentication challenge. New in 10.7.'''
        self.log('connection_willSendRequestForAuthenticationChallenge_')
        self.handleChallenge_withCompletionHandler_(challenge, None)

    def URLSession_task_didReceiveChallenge_completionHandler_(
            self, _session, _task, challenge, completionHandler):
        '''NSURLSessionTaskDelegate method'''
        if CALLBACK_HELPER_AVAILABLE:
            completionHandler.__block_signature__ = objc_method_signature(b'v@i@')
        self.log('URLSession_task_didReceiveChallenge_completionHandler_')
        self.handleChallenge_withCompletionHandler_(
            challenge, completionHandler)

    def connection_didReceiveAuthenticationChallenge_(
            self, _connection, challenge):
        '''NSURLConnection delegate method
        Sent when a connection must authenticate a challenge in order to
        download its request. Deprecated in 10.10'''
        self.log('connection_didReceiveAuthenticationChallenge_')
        self.handleChallenge_withCompletionHandler_(challenge, None)

    def handleReceivedData_(self, data):
        '''Handle received data'''
        if self.destination:
            self.destination.write(data)
        else:
            try:
                self.log(str(data))
            except Exception:
                pass
        self.bytesReceived += len(data)
        if self.expectedLength != NSURLResponseUnknownLength:
            # pylint: disable=old-division
            self.percentComplete = int(
                float(self.bytesReceived)/float(self.expectedLength) * 100.0)
            # pylint: enable=old-division

    def URLSession_dataTask_didReceiveData_(self, _session, _task, data):
        '''NSURLSessionDataDelegate method'''
        self.handleReceivedData_(data)

    def connection_didReceiveData_(self, _connection, data):
        '''NSURLConnectionDataDelegate method
        Sent as a connection loads data incrementally'''
        self.handleReceivedData_(data)


if __name__ == '__main__':
    print('This is a library of support tools for the Munki Suite.')



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""uapiobjects.py

Classes representing JSS database objects and their UAPI endpoints
"""
from .uapiobject import UAPIObject, UAPIContainer

__all__ = 'AdvancedMobileDeviceSearch', 'AlertNotification', 'Building', 'Cache', 'Category', 'ClientCheckIn', \
          'Department', 'DeviceEnrollment', 'Ebook', 'Engage', 'EnrollmentHistory', 'EnrollmentSetting', \
          'EnrollmentSettingSetting', 'Lobby', \
          'MobileDevice', 'PatchPolicy', 'ReEnrollmentSetting', 'Script', 'SelfServiceBrandingConfiguration', \
          'SelfServiceSettings', 'SSOCertificate', 'SSOSetting', 'Site', 'StartupStatus', 'SystemInformation', \
          'SystemInitialize', 'User', 'VPPAdminAccount', 'VPPSubscription'


class AdvancedMobileDeviceSearch(UAPIContainer):
    _endpoint_path = "devices/advancedSearches"


class AlertNotification(UAPIContainer):
    _endpoint_path = "notifications/alerts"
    can_put = False
    can_post = False


class APIIntegration(UAPIContainer):
    _endpoint_path = "api-integrations"


class Building(UAPIContainer):
    _endpoint_path = "settings/obj/building"


class Cache(UAPIObject):
    _endpoint_path = "settings/obj/cache"
    can_post = False
    can_delete = False


class Category(UAPIContainer):
    _endpoint_path = "settings/obj/category"


class ClientCheckIn(UAPIObject):
    _endpoint_path = "settings/obj/checkIn"
    can_post = False
    can_delete = False


class Department(UAPIContainer):
    _endpoint_path = "settings/obj/department"


class DeviceEnrollment(UAPIObject):
    _endpoint_path = "device-enrollment"


class DeviceEnrollmentPublicKey(UAPIObject):
    _endpoint_path = "device-enrollment/public-key"


class Ebook(UAPIContainer):
    _endpoint_path = "deployable/obj/ebook"
    can_put = False
    can_post = False
    can_delete = False


class Engage(UAPIObject):
    _endpoint_path = "engage"
    can_post = False
    can_delete = False


class EnrollmentHistory(UAPIContainer):
    _endpoint_path = "settings/obj/enrollment/history"
    can_post = False
    can_put = False
    can_delete = False


class EnrollmentSetting(UAPIObject):
    _endpoint_path = "settings/enrollment"  # was: settings/obj/enrollment until about 10.9
    can_post = False
    can_delete = False


# Unbelievably, this has a different schema but is intended for the same usage as EnrollmentSetting
class EnrollmentSettingSetting(UAPIObject):
    _endpoint_path = "settings/enrollment/settings"  # was: settings/obj/enrollment/settings until about 10.9
    can_post = False
    can_delete = False


class InventoryPreload(UAPIContainer):
    _endpoint_path = "inventory-preload"


class JAMFProServerURL(UAPIObject):
    _endpoint_path = "v1/jamf-pro-server-url"


class LDAPGroup(UAPIContainer):
    _endpoint_path = "ldap/groups"
    can_delete = False
    can_post = False
    can_put = False


class LDAPServer(UAPIContainer):
    _endpoint_path = "ldap/servers"
    can_delete = False
    can_post = False
    can_put = False


class Lobby(UAPIObject):
    _endpoint_path = ""
    can_put = False
    can_post = False
    can_delete = False


class MobileDevice(UAPIObject):
    _endpoint_path = "inventory/obj/mobileDevice"
    can_put = False
    can_post = False
    can_delete = False


class MobileDeviceExtensionAttribute(UAPIContainer):
    _endpoint_path = "devices/extensionAttributes"
    can_put = False
    can_post = False
    can_delete = False


class MobileDevicePrestage(UAPIContainer):
    _endpoint_path = "v1/mobile-device-prestages"
    can_put = False
    can_post = False
    can_delete = False


class PatchPolicy(UAPIContainer):
    # _endpoint_path = "patch/obj/policy"
    _endpoint_path = "patch/patch-policies"
    can_put = False
    can_post = False
    can_delete = False


# class PatchPolicyLog(UAPIContainer):
#     _endpoint_path = "patch/patch-policies/id/logs"
#     can_put = False
#     can_post = False
#     can_delete = False

class Patch(UAPIContainer):
    _endpoint_path = "patch/obj/policy"


class ReEnrollmentSetting(UAPIObject):
    _endpoint_path = "settings/obj/reenrollment"


class Script(UAPIContainer):
    _endpoint_path = "settings/scripts"


class SelfServiceBrandingConfiguration(UAPIContainer):
    _endpoint_path = "self-service/branding/configurations"


class SelfServiceSettings(UAPIObject):
    _endpoint_path = "settings/obj/selfservice"
    can_post = False
    can_delete = False


class Site(UAPIContainer):
    _endpoint_path = "settings/sites"
    can_delete = False
    can_post = False
    can_put = False


class SSOCertificate(UAPIObject):
    _endpoint_path = "settings/sso/cert"


class SSOSetting(UAPIObject):
    _endpoint_path = "settings/sso"
    can_post = False
    can_delete = False


class StartupStatus(UAPIObject):
    _endpoint_path = "startup-status"
    can_put = False
    can_post = False
    can_delete = False


class SystemInformation(UAPIObject):
    _endpoint_path = "system/obj/info"
    can_put = False
    can_post = False
    can_delete = False


class SystemInitialize(UAPIObject):
    _endpoint_path = "system/initialize"
    can_get = False
    can_put = False
    can_delete = False
    can_post = True


class User(UAPIContainer):
    _endpoint_path = "user"
    can_delete = False
    can_post = False
    can_put = False


class UserAccountSetting(UAPIContainer):
    _endpoint_path = "user/obj/preference"
    can_post = False


class VPPAdminAccount(UAPIContainer):
    _endpoint_path = "vpp/admin-accounts"
    can_put = False
    can_post = False
    can_delete = False


class VPPSubscription(UAPIContainer):
    _endpoint_path = "vpp/subscriptions"
    can_put = False
    can_post = False
    can_delete = False



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""jssobjects.py

Classes representing JSS database objects and their API endpoints
"""


from __future__ import absolute_import
import mimetypes
import os
import sys
from xml.etree import ElementTree
from xml.sax.saxutils import escape

sys.path.insert(0, '/Library/AutoPkg/JSSImporter')
import requests

from .queryset import QuerySet
from .exceptions import GetError
from .jssobject import Container, Group, JSSObject
from .tools import error_handler


__all__ = (
    'Account', 'AccountGroup', 'ActivationCode', 'AdvancedComputerSearch',
    'AdvancedMobileDeviceSearch', 'AdvancedUserSearch',
    'AllowedFileExtension', 'Building', 'BYOProfile', 'Category', 'Class',
    'Computer', 'ComputerApplication', 'ComputerApplicationUsage',
    'ComputerCheckIn', 'ComputerCommand', 'ComputerConfiguration',
    'ComputerExtensionAttribute', 'ComputerGroup',
    'ComputerHardwareSoftwareReport', 'ComputerHistory',
    'ComputerInventoryCollection', 'ComputerInvitation', 'ComputerManagement',
    'ComputerReport', 'Department', 'DirectoryBinding',
    'DiskEncryptionConfiguration', 'DistributionPoint', 'DockItem', 'EBook',
    'GSXConnection', 'HealthcareListener', 'HealthcareListenerRule', 'IBeacon',
    'InfrastructureManager', 'JSSUser', 'JSONWebTokenConfigurations',
    'LDAPServer', 'LicensedSoftware', 'MacApplication',
    'ManagedPreferenceProfile', 'MobileDevice', 'MobileDeviceApplication',
    'MobileDeviceCommand', 'MobileDeviceConfigurationProfile',
    'MobileDeviceEnrollmentProfile', 'MobileDeviceExtensionAttribute',
    'MobileDeviceGroup', 'MobileDeviceHistory', 'MobileDeviceInvitation',
    'MobileDeviceProvisioningProfile', 'NetbootServer', 'NetworkSegment',
    'OSXConfigurationProfile', 'Package', 'Patch', 'PatchAvailableTitle',
    'PatchExternalSource', 'PatchInternalSource', 'PatchPolicy',
    'PatchReport', 'PatchSoftwareTitle', 'Peripheral',
    'PeripheralType', 'Policy', 'Printer', 'RestrictedSoftware',
    'RemovableMACAddress', 'SavedSearch', 'Script', 'Site',
    'SoftwareUpdateServer', 'SMTPServer', 'UserExtensionAttribute', 'User',
    'UserGroup', 'VPPAccount', 'VPPAssignment', 'VPPInvitation', 'Webhook')

# Map Python 2 basestring type for Python 3.
if sys.version_info.major == 3:
    basestring = str

# pylint: disable=missing-docstring
class Account(Container):
    """JSS account."""
    _endpoint_path = "accounts"
    # TODO: This is pending removal.
    container = "users"
    id_url = "userid"
    search_types = {"userid": "userid", "username": "username",
                    "name": "username"}


class AccountGroup(Container):
    """Account groups are groups of users on the JSS.

    Within the API hierarchy they are actually part of accounts, but I
    seperated them.
    """

    _endpoint_path = "accounts"
    # TODO: This is pending removal.
    container = "groups"
    id_url = "groupid"
    search_types = {"groupid": "groupid", "groupname": "groupname",
                    "name": "groupname"}


class ActivationCode(JSSObject):
    _endpoint_path = "activationcode"
    can_delete = False
    can_post = False


class AdvancedComputerSearch(Container):
    _endpoint_path = "advancedcomputersearches"


class AdvancedMobileDeviceSearch(Container):
    _endpoint_path = "advancedmobiledevicesearches"


class AdvancedUserSearch(Container):
    _endpoint_path = "advancedusersearches"


class AllowedFileExtension(Container):
    _endpoint_path = "allowedfileextensions"
    can_put = False
    default_search = "extension"
    search_types = {"extension": "extension"}


class Building(Container):
    _endpoint_path = "buildings"
    root_tag = "building"


class BYOProfile(Container):
    _endpoint_path = "byoprofiles"
    root_tag = "byoprofiles"
    can_delete = False
    can_post = False
    search_types = {"sitename": "site/name", "siteid": "site/id",
                    "name": "name"}


class Category(Container):
    _endpoint_path = "categories"
    root_tag = "category"


class Class(Container):
    _endpoint_path = "classes"


class Computer(Container):
    root_tag = "computer"
    _endpoint_path = "computers"
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macaddress", "match": "match"}
    # The '/computers/match/name/{matchname}' variant is not supported
    # here because in testing, it didn't actually do anything. - It does now, at least in 10.5 -mo
    allowed_kwargs = ('subset', 'match')
    data_keys = {
        "general": {
            "mac_address": None,
            "alt_mac_address": None,
            "ip_address": None,
            "last_reported_ip": None,
            "serial_number": None,
            "udid": None,
            "jamf_version": None,
            "platform": None,
            "report_date": None,
            "initial_entry_date": None,
            "mdm_capable": None,
        },
        "location": {
            "username": None,
            "realname": None,
            "real_name": None,
            "email_address": None,
            "position": None,
            "phone": None,
            "phone_number": None,
        }
    }

    @property
    def mac_addresses(self):
        """Return a list of mac addresses for this device.

        Computers don't tell you which network device is which.
        """
        mac_addresses = [self.findtext("general/mac_address")]
        if self.findtext("general/alt_mac_address"):
            mac_addresses.append(self.findtext("general/alt_mac_address"))
            return mac_addresses


class ComputerApplication(Container):
    _endpoint_path = "computerapplications"
    can_delete = False
    can_put = False
    can_post = False
    default_search = "application"
    search_types = {"application": "application"}
    allowed_kwargs = ("version", "inventory")


class ComputerApplicationUsage(Container):
    _endpoint_path = "computerapplicationusage"
    can_delete = False
    can_put = False
    can_post = False
    allowed_kwargs = ('start_date', 'end_date')
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macaddress",
                    "match": "match"}

    @classmethod
    def _handle_kwargs(cls, kwargs):
        """Do nothing. Can be overriden by classes which need it."""
        if not all(key in kwargs for key in ('start_date', 'end_date')):
            raise TypeError(
                "This class requires a `start_date` and an `end_date` "
                "parameter.")

        # The current `build_query` implementation needs dates to be a
        # single item in the keywords dict, so combine them.
        start, end = kwargs.pop('start_date'), kwargs.pop('end_date')
        kwargs['date_range'] = (start, end)
        return kwargs


class ComputerCheckIn(JSSObject):
    _endpoint_path = "computercheckin"
    can_delete = False
    can_post = False


class ComputerCommand(Container):
    _url = "/computercommands"
    can_delete = False
    can_put = False
    # TODO: You _can_ POST computer commands, but it is not yet
    # implemented
    can_post = False


class ComputerConfiguration(Container):
    _endpoint_path = "computerconfigurations"
    root_tag = "computer_configuration"


class ComputerExtensionAttribute(Container):
    _endpoint_path = "computerextensionattributes"
    search_types = {"name": "name"}
    data_keys = {
        "description": "",
        "data_type": None,
        "input_type": {
            "type": None,  # script, Pop-up Menu, LDAP Attribute Mapping, Text Field
            "script": None,
            "popup_choices": None,  # Array of <choice>value</choice> for Pop-Up type
        },
        "inventory_display": None,
        "recon_display": None,
    }


class ComputerGroup(Group):
    _endpoint_path = "computergroups"
    root_tag = "computer_group"
    data_keys = {
        "is_smart": False,
        "criteria": None,
        "computers": None,}

    def add_computer(self, computer):
        """Add a computer to the group.

        Args:
            computer: A Computer object to add to the group.
        """
        super(ComputerGroup, self).add_device(computer, "computers")

    def remove_computer(self, computer):
        """Remove a computer from the group.

        Args:
            computer: A Computer object to add to the group.
        """
        super(ComputerGroup, self).remove_object_from_list(
            computer, "computers")


class ComputerHardwareSoftwareReport(Container):
    _endpoint_path = "computers"
    can_put = False
    can_post = False
    can_delete = False
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macadress",
                    "match": "match"}
    allowed_kwargs = ('start_date', 'end_date', 'subset')


class ComputerHardwareSoftwareReport(Container):
    """Unimplemented at this time."""
    _endpoint_path = "computerhardwaresoftwarereports"
    can_delete = False
    can_put = False
    can_post = False


class ComputerHistory(Container):
    _endpoint_path = "computerhistory"
    can_delete = False
    can_put = False
    can_post = False
    allowed_kwargs = ('subset',)
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macadress"}


class ComputerInventoryCollection(JSSObject):
    _endpoint_path = "computerinventorycollection"
    can_post = False
    can_delete = False


class ComputerInvitation(Container):
    _endpoint_path = "computerinvitations"
    can_put = False
    search_types = {"name": "name", "invitation": "invitation"}


class ComputerManagement(Container):
    _endpoint_path = "computermanagement"
    can_put = False
    can_post = False
    can_delete = False
    allowed_kwargs = ('patchfilter', 'username', 'subset')
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macadress"}


class ComputerReport(Container):
    _endpoint_path = "computerreports"
    can_put = False
    can_post = False
    can_delete = False


class Department(Container):
    _endpoint_path = "departments"
    root_tag = "department"


class DirectoryBinding(Container):
    _endpoint_path = "directorybindings"


class DiskEncryptionConfiguration(Container):
    _endpoint_path = "diskencryptionconfigurations"


class DistributionPoint(Container):
    _endpoint_path = "distributionpoints"
    root_tag = "distribution_point"


class DockItem(Container):
    _endpoint_path = "dockitems"


class EBook(Container):
    _endpoint_path = "ebooks"
    allowed_kwargs = ('subset',)


class GSXConnection(JSSObject):
    _endpoint_path = "gsxconnection"
    can_post = False
    can_delete = False


class HealthcareListener(Container):
    _endpoint_path = "healthcarelistener"
    can_post = False
    can_delete = False
    default_search = "id"
    search_types = {"id": "id"}


class HealthcareListenerRule(Container):
    _endpoint_path = "healthcarelistenerrule"
    can_delete = False
    default_search = "id"
    search_types = {"id": "id"}


class IBeacon(Container):
    _endpoint_path = "ibeacons"
    root_tag = "ibeacon"


class InfrastructureManager(Container):
    _endpoint_path = "infrastructuremanager"
    can_post = False
    can_delete = False
    default_search = "id"
    search_types = {"id": "id"}


class JSSUser(JSSObject):
    """JSSUser is deprecated."""
    _endpoint_path = "jssuser"
    can_post = False
    can_put = False
    can_delete = False


class JSONWebTokenConfigurations(JSSObject):
    _endpoint_path = "jsonwebtokenconfigurations"
    default_search = "id"
    search_types = {"id": "id"}


class LDAPServer(Container):
    _endpoint_path = "ldapservers"
    root_tag = "ldap_server"

    def search_users(self, user):
        """Search for LDAP users.

        Args:
            user: User to search for. It is not entirely clear how the
                JSS determines the results- are regexes allowed, or
                globbing?

        Returns:
            LDAPUsersResult object.

        Raises:
            Will raise a GetError if no results are found.
        """
        user_url = "%s/%s/%s" % (self.url, "user", user)
        response = self.jss.get(user_url)
        return LDAPUsersResults(self.jss, response)

    def search_groups(self, group):
        """Search for LDAP groups.

        Args:
            group: Group to search for. It is not entirely clear how the
                JSS determines the results- are regexes allowed, or
                globbing?

        Returns:
            LDAPGroupsResult object.

        Raises:
            GetError if no results are found.
        """
        group_url = "%s/%s/%s" % (self.url, "group", group)
        response = self.jss.get(group_url)
        return LDAPGroupsResults(self.jss, response)

    def is_user_in_group(self, user, group):
        """Test for whether a user is in a group.

        There is also the ability in the API to test for whether
        multiple users are members of an LDAP group, but you should just
        call is_user_in_group over an enumerated list of users.

        Args:
            user: String username.
            group: String group name.

        Returns bool.
        """
        search_url = "%s/%s/%s/%s/%s" % (self.url, "group", group,
                                         "user", user)
        response = self.jss.get(search_url)
        # Sanity check
        length = len(response)
        result = False
        if length == 1:
            # User doesn't exist. Use default False value.
            pass
        elif length == 2:
            if response.findtext("ldap_user/username") == user:
                if response.findtext("ldap_user/is_member") == "Yes":
                    result = True
        elif len(response) >= 2:
            raise GetError("Unexpected response.")
        return result

    @property
    def id(self):
        """Return object ID or None."""
        # LDAPServer's ID is in "connection"
        result = self.findtext("connection/id")
        return result or "0"

    @property
    def name(self):
        """Return object name or None."""
        # LDAPServer's name is in "connection"
        result = self.findtext("connection/name")
        return result


class LDAPUsersResults(Container):
    """Helper class for results of LDAPServer queries for users."""
    can_get = False
    can_post = False
    can_put = False
    can_delete = False


class LDAPGroupsResults(Container):
    """Helper class for results of LDAPServer queries for groups."""
    can_get = False
    can_post = False
    can_put = False
    can_delete = False


class LicensedSoftware(Container):
    _endpoint_path = "licensedsoftware"


class MacApplication(Container):
    _endpoint_path = "macapplications"
    root_tag = "mac_application"
    allowed_kwargs = ('subset',)


class ManagedPreferenceProfile(Container):
    _endpoint_path = "managedpreferenceprofiles"
    allowed_kwargs = ('subset',)


class MobileDevice(Container):
    """Mobile Device objects include a "match" search type which queries
    across multiple properties.
    """

    _endpoint_path = "mobiledevices"
    root_tag = "mobile_device"
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macadress",
                    "match": "match"}
    allowed_kwargs = ('subset',)


class MobileDeviceApplication(Container):
    _endpoint_path = "mobiledeviceapplications"
    allowed_kwargs = ('subset',)


class MobileDeviceCommand(Container):
    _endpoint_path = "mobiledevicecommands"
    can_put = False
    can_delete = False
    search_types = {"name": "name", "uuid": "uuid",
                    "command": "command"}
    # TODO: This object _can_ post, but it works a little differently
    # and is not yet implemented
    can_post = False


class MobileDeviceConfigurationProfile(Container):
    _endpoint_path = "mobiledeviceconfigurationprofiles"
    allowed_kwargs = ('subset',)


class MobileDeviceEnrollmentProfile(Container):
    _endpoint_path = "mobiledeviceenrollmentprofiles"
    search_types = {"name": "name", "invitation": "invitation"}
    allowed_kwargs = ('subset',)


class MobileDeviceExtensionAttribute(Container):
    _endpoint_path = "mobiledeviceextensionattributes"


class MobileDeviceGroup(Group):
    _endpoint_path = "mobiledevicegroups"
    root_tag = "mobile_device_group"

    def add_mobile_device(self, device):
        """Add a mobile_device to the group.

        Args:
            device: A MobileDevice object to add to group.
        """
        super(MobileDeviceGroup, self).add_device(device, "mobile_devices")

    def remove_mobile_device(self, device):
        """Remove a mobile_device from the group.

        Args:
            device: A MobileDevice object to remove from the group.
        """
        super(MobileDeviceGroup, self).remove_object_from_list(
            device, "mobile_devices")


class MobileDeviceHistory(Container):
    _endpoint_path = "mobiledevicehistory"
    can_delete = False
    can_put = False
    can_post = False
    allowed_kwargs = ('subset',)
    search_types = {"name": "name", "serial_number": "serialnumber",
                    "udid": "udid", "macaddress": "macadress"}


class MobileDeviceInvitation(Container):
    _endpoint_path = "mobiledeviceinvitations"
    can_put = False
    search_types = {"invitation": "invitation"}


class MobileDeviceProvisioningProfile(Container):
    _endpoint_path = "mobiledeviceprovisioningprofiles"
    search_types = {"name": "name", "uuid": "uuid"}
    allowed_kwargs = ('subset',)


class NetbootServer(Container):
    _endpoint_path = "netbootservers"


class NetworkSegment(Container):
    _endpoint_path = "networksegments"
    root_tag = "network_segment"
    data_keys = {
        "starting_address": "",
        "ending_address": "",
        "distribution_server": None,
        "distribution_point": None,
        "url": None,
        "netboot_server": None,
        "swu_server": None,
        "building": None,
        "department": None,
        "override_buildings": False,
        "override_departments": False,
    }


class OSXConfigurationProfile(Container):
    _endpoint_path = "osxconfigurationprofiles"
    root_tag = "os_x_configuration_profile"
    search_types = {"name": "name"}
    allowed_kwargs = ('subset',)
    _name_element = "general/name"
    data_keys = {
        "general": {
            "description": "",
            "category": "",
            "distribution_method": "Install Automatically",
            "user_removable": "true",
            "level": "computer",
            "uuid": None,
            "redeploy_on_update": "Newly Assigned",
            "payloads": None,
        },
        "scope": {
            "computers": None,
            "computer_groups": None,
            "buildings": None,
            "departments": None,
            "exclusions": {
                "computers": None,
                "computer_groups": None,
                "buildings": None,
                "departments": None,
            },
        }
        # TODO: Self Service
    }

    def set_category(self, category):
        """Set the OSXConfigurationProfile's category.

        Args:
            category: A category object.
        """
        pcategory = self.find("general/category")
        pcategory.clear()

        if isinstance(category, Category):
            id_ = ElementTree.SubElement(pcategory, "id")
            id_.text = category.id
            name = ElementTree.SubElement(pcategory, "name")
            name.text = category.name
        elif isinstance(category, basestring):
            name = ElementTree.SubElement(pcategory, "name")
            name.text = category

    def add_payloads(self, payloads_contents):
        """Add xml configuration profile to the correct tag in the OSXConfigurationProfile object.

        The profile content will be XML encoded prior to addition,
        so there's no need to encode prior to addition with this
        method.

        Args:
            payloads_contents (str, unicode): Mobile Configuration Profile.
        """
        payloads_contents_tag = self.find("general/payloads")
        if not payloads_contents_tag:
            payloads_contents_tag = ElementTree.SubElement(
                self.find("general"), "payloads")

        # Normally, when embedding xml within xml you would think that a normal person would
        # require escaping to avoid parsing issues. But nope, we just put the profile inline with the
        # payloads tag.
        payloads_contents_tag.text = payloads_contents

        uuid_tag = self.find("uuid")
        if not uuid_tag:
            uuid_tag = ElementTree.SubElement(
                self.find("general"), "uuid")
        uuid_tag.text = "336001A6-460A-4213-B000-ECAE32E8429E"


class Package(Container):
    _endpoint_path = "packages"
    root_tag = "package"
    data_keys = {
        "category": None,
        "info": None,
        "notes": None,
        "priority": "10",
        "reboot_required": "false",
        "fill_user_template": "false",
        "fill_existing_users": "false",
        "boot_volume_required": "true",
        "allow_uninstalled": "false",
        "os_requirements": None,
        "required_processor": "None",   # Really. The string "None".
        "switch_with_package": "Do Not Install",
        "install_if_reported_available": "false",
        "reinstall_option": "Do Not Reinstall",
        "triggering_files": None,
        "send_notification": "false",}

    def _new(self, name, **kwargs):
        """Create a new Package from scratch.

        Args:
            name: String filename of the package to use for the
                Package object's Display Name (here, "name").
                Will also be used as the "filename" value. Casper will
                let you specify different values, but it is not
                recommended.
            kwargs:
                Accepted keyword args include all top-level keys.
                Values will be cast to string. (Int 10, bool False
                become string values "10" and "false").
        """
        # We want these to match, so circumvent the for loop.
        # ElementTree.SubElement(self, "name").text = name
        super(Package, self)._new(name, **kwargs)
        ElementTree.SubElement(self, "filename").text = name

    def set_os_requirements(self, requirements):
        """Set package OS Requirements

        Args:
            requirements: A string of comma seperated OS versions. A
                lowercase "x" is allowed as a wildcard, e.g.  "10.9.x"
        """
        self.find("os_requirements").text = requirements

    def set_category(self, category):
        """Set package category

        Args:
            category: String of an existing category's name, or a
                Category object.
        """
        # For some reason, packages only have the category name, not the
        # ID.
        if isinstance(category, Category):
            name = category.name
        else:
            name = category
        self.find("category").text = name


# DEPRECATED
class Patch(Container):
    _endpoint_path = "patches"
    root_tag = "software_title"
    can_post = False
    allowed_kwargs = ('subset',)
    # The /patches/id/{id}/version/{version} variant is not currently
    # implemented.


class PatchAvailableTitle(Container):
    _endpoint_path = "patchavailabletitles"
    can_delete = False
    can_post = False
    can_put = False
    search_types = {"sourceid": "sourceid"}
    default_search = "sourceid"


class PatchExternalSource(Container):
    _endpoint_path = "patchexternalsources"
    root_tag = "patch_external_source"
    data_keys = {
        'host_name': None,
        'ssl_enabled': 'true',
        'port': '443',
    }


class PatchInternalSource(Container):
    _endpoint_path = "patchinternalsources"
    can_delete = False
    can_put = False
    can_post = False


class PatchReport(Container):
    _endpoint_path = "patchreports"
    can_delete = False
    can_put = False
    can_post = False
    default_search = "patchsoftwaretitleid"
    search_types = {"patchsoftwaretitleid": "patchsoftwaretitleid"}


class PatchSoftwareTitle(Container):
    _endpoint_path = "patchsoftwaretitles"
    root_tag = "patch_software_title"
    data_keys = {
        "name_id": None,
        "source_id": None,
        "notifications": {
            "web_notification": "true",
            "email_notification": "false"
        },
        "versions": None
    }

    def add_package(self, pkg, version):
        """Add a Package object to the software title.

        Args:
            pkg: A Package object to add.
            version: The version of the package being added.
        """
        if isinstance(pkg, Package):
            version_el = ElementTree.SubElement(self.find('versions'), 'version')
            software_version = ElementTree.SubElement(version_el, 'software_version')
            software_version.text = version
            package_version = self.add_object_to_path(pkg, version_el)

        else:
            raise ValueError("Please pass a Package object to parameter: "
                             "pkg.")


class PatchPolicy(Container):
    _endpoint_path = "patchpolicies"
    root_tag = "patch_policy"
    search_types = {"id": "id", "softwaretitleconfigid": "softwaretitleconfigid/id"}
    allowed_kwargs = ('subset',)
    _name_element = "general/name"
    data_keys = {
        "general": {
            "enabled": "true",
            "target_version": None,
            "release_date": None,
            "incremental_updates": "false",
            "reboot": "false",
            "minimum_os": None,
            "kill_apps": None,
            "distribution_method": "selfservice",
            "allow_downgrade": "true",
            "patch_unknown": "true"
        },
        "scope": {
            "computers": None,
            "computer_groups": None,
            "buildings": None,
            "departments": None,
            "limitations": {
                "network_segments": None,
                "ibeacons": None
            },
            "exclusions": {
                "computers": None,
                "computer_groups": None,
                "buildings": None,
                "departments": None,
                "network_segments": None,
                "ibeacons": None,
            }
        },
        "user_interaction": {
            "install_button_text": "Update",
            "self_service_description": None,
            "notifications": {
                "notification_enabled": "true",
                "notification_type": "Self Service",
                "notification_subject": "Update Available",
                "notification_message": "Update Available",
                "reminders": {
                    "notification_reminders_enabled": "true",
                    "notification_reminder_frequency": "1"
                }
            }
        }
    }

    def add_object_to_scope(self, obj):
        """Add an object to the appropriate scope block.

        Args:
            obj: JSSObject to add to scope. Accepted subclasses are:
                Computer
                ComputerGroup
                Building
                Department

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, Computer):
            self.add_object_to_path(obj, "scope/computers")
        elif isinstance(obj, ComputerGroup):
            self.add_object_to_path(obj, "scope/computer_groups")
        elif isinstance(obj, Building):
            self.add_object_to_path(obj, "scope/buildings")
        elif isinstance(obj, Department):
            self.add_object_to_path(obj, "scope/departments")
        else:
            raise TypeError

    def clear_scope(self):
        """Clear all objects from the scope, including exclusions."""
        clear_list = ["computers", "computer_groups", "buildings",
                      "departments", "limit_to_users/user_groups",
                      "limitations/users", "limitations/user_groups",
                      "limitations/network_segments", "exclusions/computers",
                      "exclusions/computer_groups", "exclusions/buildings",
                      "exclusions/departments", "exclusions/users",
                      "exclusions/user_groups", "exclusions/network_segments"]
        for section in clear_list:
            self.clear_list("%s%s" % ("scope/", section))

    def add_object_to_exclusions(self, obj):
        """Add an object to the appropriate scope exclusions
        block.

        Args:
            obj: JSSObject to add to exclusions. Accepted subclasses
                    are:
                Computer
                ComputerGroup
                Building
                Department

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, Computer):
            self.add_object_to_path(obj, "scope/exclusions/computers")
        elif isinstance(obj, ComputerGroup):
            self.add_object_to_path(obj, "scope/exclusions/computer_groups")
        elif isinstance(obj, Building):
            self.add_object_to_path(obj, "scope/exclusions/buildings")
        elif isinstance(obj, Department):
            self.add_object_to_path(obj, "scope/exclusions/departments")
        else:
            raise TypeError

    def add_object_to_limitations(self, obj):
        """Add an object to the appropriate scope limitations
        block.

        Args:
            obj: JSSObject to add to limitations. Accepted subclasses
                are:
                    User
                    UserGroup
                    NetworkSegment
                    IBeacon

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, User):
            self.add_object_to_path(obj, "scope/limitations/users")
        elif isinstance(obj, UserGroup):
            self.add_object_to_path(obj, "scope/limitations/user_groups")
        elif isinstance(obj, NetworkSegment):
            self.add_object_to_path(obj, "scope/limitations/network_segments")
        elif isinstance(obj, IBeacon):
            self.add_object_to_path(obj, "scope/limitations/ibeacons")
        else:
            raise TypeError



class Peripheral(Container):
    _endpoint_path = "peripherals"
    search_types = {}
    allowed_kwargs = ('subset',)


class PeripheralType(Container):
    _endpoint_path = "peripheraltypes"
    search_types = {}


# pylint: disable=too-many-instance-attributes
# This class has a lot of convenience attributes. Sorry pylint.
class Policy(Container):
    _endpoint_path = "policies"
    root_tag = "policy"
    search_types = {"name": "name", "category": "category"}
    allowed_kwargs = ('subset',)
    _name_element = "general/name"
    data_keys = {
        "general": {
            "enabled": "true",
            "frequency": "Once per computer",
            "category": "",},
        "scope": {
            "computers": None,
            "computer_groups": None,
            "buildings": None,
            "departments": None,
            "exclusions": {
                "computers": None,
                "computer_groups": None,
                "buildings": None,
                "departments": None,},},
        "self_service": {
            "use_for_self_service": "true"},
        "package_configuration": {
            "packages": None},
        "scripts": None,
        "maintenance": {
            "recon": "true"},
    }

    def add_object_to_scope(self, obj):
        """Add an object to the appropriate scope block.

        Args:
            obj: JSSObject to add to scope. Accepted subclasses are:
                Computer
                ComputerGroup
                Building
                Department

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, Computer):
            self.add_object_to_path(obj, "scope/computers")
        elif isinstance(obj, ComputerGroup):
            self.add_object_to_path(obj, "scope/computer_groups")
        elif isinstance(obj, Building):
            self.add_object_to_path(obj, "scope/buildings")
        elif isinstance(obj, Department):
            self.add_object_to_path(obj, "scope/departments")
        else:
            raise TypeError

    def clear_scope(self):
        """Clear all objects from the scope, including exclusions."""
        clear_list = ["computers", "computer_groups", "buildings",
                      "departments", "limit_to_users/user_groups",
                      "limitations/users", "limitations/user_groups",
                      "limitations/network_segments", "exclusions/computers",
                      "exclusions/computer_groups", "exclusions/buildings",
                      "exclusions/departments", "exclusions/users",
                      "exclusions/user_groups", "exclusions/network_segments"]
        for section in clear_list:
            self.clear_list("%s%s" % ("scope/", section))

    def add_object_to_exclusions(self, obj):
        """Add an object to the appropriate scope exclusions
        block.

        Args:
            obj: JSSObject to add to exclusions. Accepted subclasses
                    are:
                Computer
                ComputerGroup
                Building
                Department

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, Computer):
            self.add_object_to_path(obj, "scope/exclusions/computers")
        elif isinstance(obj, ComputerGroup):
            self.add_object_to_path(obj, "scope/exclusions/computer_groups")
        elif isinstance(obj, Building):
            self.add_object_to_path(obj, "scope/exclusions/buildings")
        elif isinstance(obj, Department):
            self.add_object_to_path(obj, "scope/exclusions/departments")
        else:
            raise TypeError

    def add_object_to_limitations(self, obj):
        """Add an object to the appropriate scope limitations
        block.

        Args:
            obj: JSSObject to add to limitations. Accepted subclasses
                are:
                    User
                    UserGroup
                    NetworkSegment
                    IBeacon

        Raises:
            TypeError if invalid obj type is provided.
        """
        if isinstance(obj, User):
            self.add_object_to_path(obj, "scope/limitations/users")
        elif isinstance(obj, UserGroup):
            self.add_object_to_path(obj, "scope/limitations/user_groups")
        elif isinstance(obj, NetworkSegment):
            self.add_object_to_path(obj, "scope/limitations/network_segments")
        elif isinstance(obj, IBeacon):
            self.add_object_to_path(obj, "scope/limitations/ibeacons")
        else:
            raise TypeError

    def add_package(self, pkg, action_type="Install"):
        """Add a Package object to the policy with action=install.

        Args:
            pkg: A Package object to add.
            action_type (str, optional): One of "Install", "Cache", or
                "Install Cached".  Defaults to "Install".
        """
        if isinstance(pkg, Package):
            if action_type not in ("Install", "Cache", "Install Cached"):
                raise ValueError
            package = self.add_object_to_path(
                pkg, "package_configuration/packages")
            # If there's already an action specified, get it, then
            # overwrite. Otherwise, make a new subelement.
            action = package.find("action")
            if not action:
                action = ElementTree.SubElement(package, "action")
            action.text = action_type
        else:
            raise ValueError("Please pass a Package object to parameter: "
                             "pkg.")

    def remove_package(self, pkg):  # type: (Union[Package,str]) -> None
        """Remove a Package object from the policy.

        Args:
            pkg: A Package object, id or name to remove.

        Raises:
            ValueError, if object is not found or xml element cannot be found.
        """
        self.remove_object_from_list(pkg, "package_configuration/packages")

    def get_packages(self):  # type: () -> Optional[QuerySet]
        """Get all package objects associated with this policy."""
        package_set = QuerySet.from_response(
            'Package',
            self.package_configuration.packages.findall('package'),
            jss=self.jss,
        )

        return package_set

    def add_script(self, script, priority='After', parameters=None):
        """Add a Script object to the policy with priority=After.

        Args:
            script: A Script object to add.
            priority (str, optional): One of "After" or "Before".
            parameters (list): List of parameters starting from parameter 4.
        """
        if isinstance(script, Script):
            if priority not in ("After", "Before"):
                raise ValueError
            script_tag = self.add_object_to_path(
                script, "scripts")
            priority_tag = script_tag.find("priority")
            if not priority_tag:
                priority_tag = ElementTree.SubElement(script_tag, "priority")
            priority_tag.text = priority

            if isinstance(parameters, list):
                for param_idx in range(4, 11, 1):
                    param_tag = script_tag.find("parameter{}".format(param_idx))
                    if not param_tag:
                        param_tag = ElementTree.SubElement(script_tag, "parameter{}".format(param_idx))

                    if param_idx - 4 > len(parameters) - 1:
                        param_tag.text = ""
                    else:
                        param_tag.text = parameters[param_idx - 4]
        else:
            raise ValueError("Please pass a Script object to parameter: script")

    def get_scripts(self, priority=None):
        """Get Scripts that are referenced by this policy.

        Args:
            priority (str, optional): One of "After" or "Before" to filter attached scripts by their priority.
        """
        script_set = QuerySet.from_response(
            Script,
            self.scripts,
        )

        return script_set

    def set_self_service(self, state=True):
        """Set use_for_self_service to bool state."""
        self.set_bool(self.find("self_service/use_for_self_service"), state)

    def set_recon(self, state=True):
        """Set policy's recon value to bool state."""
        self.set_bool(self.find("maintenance/recon"), state)

    def set_category(self, category):
        """Set the policy's category.

        Args:
            category: A category object.
        """
        pcategory = self.find("general/category")
        pcategory.clear()

        if isinstance(category, Category):
            id_ = ElementTree.SubElement(pcategory, "id")
            id_.text = category.id
            name = ElementTree.SubElement(pcategory, "name")
            name.text = category.name
        elif isinstance(category, basestring):
            name = ElementTree.SubElement(pcategory, "name")
            name.text = category

# pylint: enable=too-many-instance-attributes, too-many-locals


class Printer(Container):
    _endpoint_path = "printers"


class RemovableMACAddress(Container):
    _endpoint_path = "removablemacaddresses"


class RestrictedSoftware(Container):
    _endpoint_path = "restrictedsoftware"


class SavedSearch(Container):
    _endpoint_path = "savedsearches"
    can_put = False
    can_post = False
    can_delete = False


class Script(Container):
    _endpoint_path = "scripts"
    root_tag = "script"

    def add_script(self, script_contents):
        """Add script code to the correct tag in the Script object.

        The script content will be XML encoded prior to addition,
        so there's no need to encode prior to addition with this
        method.

        Args:
            script_contents (str, unicode): Script code.
        """
        escaped_script_contents = escape(script_contents)
        script_contents_tag = self.find("script_contents")
        if not script_contents_tag:
            script_contents_tag = ElementTree.SubElement(
                self, "script_contents")
        script_contents_tag.text = escaped_script_contents


class Site(Container):
    _endpoint_path = "sites"
    root_tag = "site"


class SMTPServer(JSSObject):
    _endpoint_path = "smtpserver"
    can_post = False
    can_delete = False


class SoftwareUpdateServer(Container):
    _endpoint_path = "softwareupdateservers"


class UserExtensionAttribute(Container):
    _endpoint_path = "userextensionattributes"


class User(Container):
    _endpoint_path = "users"


class UserGroup(Container):
    _endpoint_path = "usergroups"


class VPPAccount(Container):
    _endpoint_path = "vppaccounts"
    root_tag = "vpp_account"


class VPPAssignment(Container):
    _endpoint_path = "vppassignments"


class VPPInvitation(Container):
    _endpoint_path = "vppinvitations"


class Webhook(Container):
    _endpoint_path = "webhooks"


# pylint: enable=missing-docstring



#!/usr/bin/python
# Copyright (C) 2015 Michael Lynn
# https://gist.github.com/pudquick/1362a8908be01e23041d
# https://michaellynn.github.io/2015/08/08/learn-you-a-better-pyobjc-bridgesupport-signature/

# This version has been modified by Shea Craig to handle High Sierra.

"""mount_shares_better.py

Mount file shares on OS X.
"""


from __future__ import absolute_import
from distutils.version import StrictVersion
import subprocess

# PyLint cannot properly find names inside Cocoa libraries, so issues bogus
# No name 'Foo' in module 'Bar' warnings. Disable them.
# pylint: disable=no-name-in-module
from CoreFoundation import CFURLCreateWithString
import Foundation  # pylint: disable=unused-import
from objc import (initFrameworkWrapper, pathForFramework, loadBundleFunctions)
# pylint: enable=no-name-in-module


class AttrDict(dict):
    """Attribute Dictionary"""
    __getattr__ = dict.__getitem__
    __setattr__ = dict.__setitem__


NetFS = AttrDict()  # pylint: disable=invalid-name
# Can cheat and provide 'None' for the identifier, it'll just use
# frameworkPath instead.
# scan_classes=False means only add the contents of this Framework.
# pylint: disable=invalid-name
NetFS_bundle = initFrameworkWrapper(
    'NetFS', frameworkIdentifier=None,
    frameworkPath=pathForFramework('NetFS.framework'), globals=NetFS,
    scan_classes=False)
# pylint: enable=invalid-name

# https://developer.apple.com/library/mac/documentation/Cocoa/Conceptual/ObjCRuntimeGuide/Articles/ocrtTypeEncodings.html
# Fix NetFSMountURLSync signature
del NetFS['NetFSMountURLSync']
loadBundleFunctions(NetFS_bundle, NetFS, [('NetFSMountURLSync', b'i@@@@@@o^@')])


def mount_share(share_path):
    """Mounts a share at /Volumes

    Args:
        share_path: String URL with all auth info to connect to file share.

    Returns:
        The mount point or raises an error.
    """
    sh_url = CFURLCreateWithString(None, share_path, None)

    # Set UI to reduced interaction
    if is_high_sierra():
        open_options = None
    else:
        open_options = {NetFS.kNAUIOptionKey: NetFS.kNAUIOptionNoUI}
    # Allow mounting sub-directories of root shares
    if is_high_sierra():
        mount_options = None
    else:
        mount_options = {NetFS.kNetFSAllowSubMountsKey: True}
    # Build our connected pointers for our results
    result, output = NetFS.NetFSMountURLSync(
        sh_url, None, None, None, open_options, mount_options, None)

    # Check if it worked
    if result != 0:
        raise Exception('Error mounting url "%s": %s' % (share_path, output))
    # Return the mountpath
    return str(output[0])


def is_high_sierra():
    version = StrictVersion(
        subprocess.check_output(['sw_vers', '-productVersion']).decode().strip())
    return version >= StrictVersion('10.13')



#!/usr/bin/env python
# Copyright (C) 2014-2017 Shea G Craig
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import
from datetime import datetime
import logging
import sys

sys.path.insert(0, '/Library/AutoPkg/JSSImporter')
import requests

logger = logging.getLogger(__name__)


class UAPIAuth(requests.auth.AuthBase):

    def __init__(self, username, password, fetch_url='/uapi/auth/tokens', token=None, expires=None):
        self.username = username
        self.password = password
        self.token = token
        self.expires = expires if expires else datetime.now()
        self.fetch_url = fetch_url

    def __call__(self, r):  # type: (requests.PreparedRequest) -> requests.PreparedRequest
        if self.expires is not None and self.expires < datetime.now():
            logger.debug("Token expiry has passed, fetching a new token.")
            self._get_token()

        r.headers['Authorization'] = 'jamf-token {}'.format(self.token)
        r.register_hook('response', self.handle_401)
        return r

    def _get_token(self):
        r = requests.post(self.fetch_url, auth=(self.username, self.password), verify=False)
        r.raise_for_status()
        data = r.json()

        self.token = data['token']
        #self.expires = datetime.utcfromtimestamp(data['expires'])

    def handle_401(self, r, **kwargs):  # type: (requests.Response, dict) -> requests.Response
        """
        Takes the given response, fetches a new token, and retries
        :param r:
        :param kwargs:
        :return:
        """
        if r.status_code is not 401:
            return r

        logger.debug("Server returned HTTP 401, getting a new token")
        self._get_token()
